#ifndef BRUSH_H
#define BRUSH_H


typedef struct
{
   int type;
/* Brushes defining map geometry */
#define BRUSH_NORMAL     1     /* a normal brush */
#define BRUSH_SUBTRACT   2     /* a subtractive brush, not currently used */
#define BRUSH_POLYGON    3     /* a lone polygon */

#define BRUSH_Q3_CURVE   4     /* a Quake 3 curve, array of bezier patches */

   const char *name;


/* Flags for this type of brush */
   int flags;
   
/* The brush shouldn't be saved */
#define BRUSH_F_NOSAVE   0x01

/* The brush is 'normal', i.e. num_edges and edges are valid, the brush is
drawn by drawing the edges, faces have valid verts structures that trace the
edge of the face, faces have texdef:s, etc. If this isn't set, there'll have
to be code specific for the type of brush somewhere to handle it. */
#define BRUSH_F_NORMAL   0x02

/* Brush can be used in CSG operations */
#define BRUSH_F_CSG      0x04


   int (*b_new)(struct brush_s *b);
   int (*b_copy)(struct brush_s *dst,const struct brush_s *src);
   void (*b_free)(struct brush_s *b);
   void (*b_update)(struct brush_s *b);
} brush_type_t;


typedef struct brush_normal_align_s
{
   vec2_t scale;
   vec2_t shift;
   scalar_t rotate;
} brush_normal_align_t;


typedef struct face_s
{
	float   dist;                   /* distance to the face's plane */
	vec3_t  normal;                 /* Normal to plane */
	vec3_t  center;
	int    *verts;                  /* indices in the brush->verts array */
	int     num_verts;

   texdef_t tex;
   
   vec2_t *texv;
   /* s and t coordinates for the vertices on this plane, in pixels */

   brush_normal_align_t a;
   /* should perhaps put this in a union, but we don't have any other
   alignment options at the moment */

   struct S_r_face_s *r;
} face_t;


/* Extra information for BRUSH_Q3_CURVE */
typedef struct
{
/* Size of bezier patch vertex array, number of patches is (dim-1)/2 */
   int sizex,sizey;
/* Texture alignment information, the s and t coordinates at vertex x,y
   are in a[x+y*sizex].x and a[x+y*sizey].y , i.e. the alignment information
   for vertex verts[i] is in a[i] */
   vec2_t *a;
} brush_q3_curve_t;


typedef struct brush_s
{
	struct brush_s  *Next;
	struct brush_s  *Last;
   struct brush_s  *VNext;
   struct brush_s  *VLast;
   map_uid_t uid;               /* Identification number, MUST be unique */
   unsigned char real;

   int misc;
   
   brush_type_t *bt;

	map_uid_t Entity;
	map_uid_t Group;

	face_t *faces;
	int     num_faces;

	vec3_t *verts;                 /* Array of vertices */
	int     num_verts;

	edge_t *edges;                 /* Array of edges */
	int     num_edges;             /* (in the local verts array) */

	vec3_t   center;

   texdef_t tex;
   /* This should always have a valid texturename, since it's used to
   display the name for this brush. For normal brushes, it'll be a copy
   of the texturename of the first plane. */

/* Extra stuff used by some types of brushes */
   union
   {
      brush_q3_curve_t q3c;
   } x;

   struct brush_s *disp;

   int drawn; /* often the only renderer field, avoid having to allocate
                 another structure */
   struct S_r_brush_s *r;
} brush_t;


brush_t *B_New(int type);
void B_Free(brush_t *b);
void B_FreeC(brush_t *b);

void B_Link(brush_t *b);
void B_Unlink(brush_t *b);

void B_ChangeType(brush_t *b,int ntype);

/*
Duplicates:
  num_faces
  face
  face.verts
  face.num_verts
  face.tex
  face.normal
  face.a

  num_verts
  verts

  num_edges
  edges

  tex

  EntityRef
  Group

  bt
*/
brush_t *B_Duplicate(brush_t *dest, const brush_t *source);

int B_Size(const brush_t *b);

void B_Update(brush_t *b);


/* Duplicates everything */
face_t *F_Copy(face_t *dst,const face_t *src);
/* Everything except vertices */
face_t *F_CopyP(face_t *dst,const face_t *src);

#endif

