/*
cmd/games/qengine/qmap.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "qmap.h"

#include "brush.h"
#include "camera.h"
#include "config.h"
#include "csg.h"
#include "entity.h"
#include "group.h"
#include "map.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"
#include "s_message.h"
#include "s_texture.h"
#include "token.h"
#include "uid.h"



static config_var_t *cv_qmap_map_format;


int (*qmap_loadtexinfo)(texdef_t *tex);
void (*qmap_savetexinfo)(texdef_t *tex,FILE *fp);


/**********
Misc. stuff
**********/

static int realtoint(float r)
{
   if ((r<0.01)&&(r>-0.01)) return 0;
   if (r<0)
   {
      r-=0.5;
      return ceil(r);
   }
   if (r>0)
   {
      r+=0.5;
      return floor(r);
   }
   return 0;
}


static float snaptoint(float f)
{
   if (fabs(f-realtoint(f))<0.001) return realtoint(f);
   return f;
}


/****************
Map loading stuff
****************/

static const char *errors[]=
{
"No error!",
"Out of memory!",
"Error parsing file!"
};


static char buf[16384]; /* TODO */

#if TODO
static char (*extra_tex)[32];
static int num_extra;

static int load_camera;
#endif

static map_uid_t mapgroup;
static map_uid_t infogroup;
static group_t *infogroup_p;


static int LoadTexInfo(texdef_t *t,brush_normal_align_t *a)
{
   GETTOKEN(0,TOKEN_ALLNAME);
   strcpy(t->name,token);

   S_T_Cache(t->name);

   GETTOKEN(0,TOKEN_NUMBER);
   a->shift.x=atof(token);
   GETTOKEN(0,TOKEN_NUMBER);
   a->shift.y=atof(token);

   GETTOKEN(0,TOKEN_NUMBER);
   a->rotate=atof(token);

   GETTOKEN(0,TOKEN_NUMBER);
   a->scale.x=atof(token);
   GETTOKEN(0,TOKEN_NUMBER);
   a->scale.y=atof(token);

   if (qmap_loadtexinfo) /* Load game-specific stuff */
      return qmap_loadtexinfo(t);
   return ERROR_NO;
}

static void SaveTexInfo(face_t *f,FILE *fp)
{
   fprintf(fp, "%s %0.0f %0.0f %0.0f %g %g",
               f->tex.name[0]?f->tex.name:DEFAULT_TEXTURE,
               f->a.shift.x,
               f->a.shift.y,
               f->a.rotate,
               f->a.scale.x,
               f->a.scale.y);
               
   if (qmap_savetexinfo) /* Save game-specific stuff */
      qmap_savetexinfo(&f->tex,fp);
}


/****************************************
Game independent map loading/saving stuff
****************************************/

static map_uid_t FindGroup(const char *name)
{
   group_t *g;

   for (g=M.GroupHead;g;g=g->Next)
   {
      if (!stricmp(name,g->fullname))
         return g->uid;
   }
   return M.WorldGroup;
}

static void Comments(char *buf)
{
#define CHK(y) !strncmp(buf,y,strlen(y))

   while (buf[strlen(buf)-1]<=32) buf[strlen(buf)-1]=0;

   // Group information,
   if (CHK("// Quest Group Name: "))
   {
      char grpname[64];
      group_t *g;
      map_uid_t parent;

      strcpy(grpname,&buf[21]);
      if (strchr(grpname,'.'))
      {
         *(strrchr(grpname,'.'))=0;
         parent=Grp_Find(grpname);
         if (!parent)
         {
            S_NewMessage("Warning: Can't find group '%s'!",grpname);
            infogroup=0;
            infogroup_p=NULL;
         }
         else
         {
            infogroup=Grp_Create(&buf[21+1+strlen(grpname)],parent);
            infogroup_p=UID_Group(infogroup);
         }
      }
      else
      {
         g=UID_Group(M.WorldGroup);
         if (!strcmp(grpname,g->groupname))
         {
            infogroup=g->uid;
            infogroup_p=g;
         }
         else
         {
            infogroup=Grp_Create(grpname,0);
            infogroup_p=UID_Group(infogroup);
         }
      }
   }
   else
   if (CHK("// Quest Group Flags: "))
   {
      if (infogroup_p)
         infogroup_p->flags=atoi(&buf[22]);
   }
   else
   if (CHK("// Quest Group Color: "))
   {
      if (infogroup_p)
         infogroup_p->color=atoi(&buf[22]);
   }
   else
   // Brush/entity groups.
   if (CHK("//Quest Brush Group: "))
   {
      mapgroup=FindGroup(&buf[21]);
   }
   else
   if (CHK("//Quest Entity Group: "))
   {
      mapgroup=FindGroup(&buf[22]);
   }
#if TODO
   else
   // Viewport stuff.
   if (CHK("// ** "))
   {
      int vport;
      char keystr[128];
      float val1,val2,val3,val4;
      viewport_t *vp;

      load_camera=1;

      sscanf(&buf[6],"%i %s %f %f %f %f\n",
         &vport,keystr,&val1,&val2,&val3,&val4);

      vp=&M.display.vport[vport];

      if (!strcmp(keystr,"num_vports"))
      {
         if (vport<MAX_NUM_VIEWPORTS)
            M.display.num_vports=vport;
         else
            M.display.num_vports=MAX_NUM_VIEWPORTS;
      }
      else
      if (!strcmp(keystr,"pos"))
      {
         vp->camera_pos.x = (int)val1;
         vp->camera_pos.y = (int)val2;
         vp->camera_pos.z = (int)val3;
      }
      else
      if (!strcmp(keystr,"dir"))
      {
         vp->camera_dir = (int)val1;
         vp->axis_aligned=1;
      }
      else
      if (strcmp(keystr, "fullbright") == 0)
      {
         vp->fullbright = (int)val1;
      }
      else
      if (strcmp(keystr, "zoom_amt") == 0)
      {
         vp->zoom_amt = val1;
      }
      else
      if (strcmp(keystr, "mode") == 0)
      {
         vp->mode = (int)val1;
      }
      else
      if (!strcmp(keystr,"spos"))
      {
         vp->f_xmin=val1;
         vp->f_ymin=val2;
         vp->f_xmax=val3;
         vp->f_ymax=val4;
      }
      else
      if (!strcmp(keystr,"angle"))
      {
         vp->axis_aligned=0;
         vp->rot_x=val1;
         vp->rot_y=val2;
         vp->rot_z=val3;
      }
   }
#endif
#if TODO
   else
   // Extra textures.
   if (CHK("// T: ") && Game.tex.cache)
   {
      if (num_extra!=-1)
      {
         char (*temp)[32];
   
         temp=Q_realloc(extra_tex,sizeof(extra_tex[0])*(num_extra+1));
         if (!temp)
         {
            HandleError("QMap_LoadMap","Out of memory loading extra textures!");
            Q_free(extra_tex);
            num_extra=-1;
         }
         else
         {
            extra_tex=temp;
            strcpy(extra_tex[num_extra],&buf[6]);
            num_extra++;
         }
      }
   }
#endif
}


static int LoadBrush_Q3C(map_uid_t ent,brush_t **res)
{
   int nx,ny;
   int i,j,k;
   
   brush_t *b;
   face_t *f;

   b=B_New(BRUSH_Q3_CURVE);
   if (!b)
      return ERROR_NOMEM;
   *res=b;
   
   EXPECT(1,TOKEN_MISC,"{");

   GETTOKEN(1,TOKEN_ALLNAME);
   strcpy(b->tex.name,token);

   S_T_Cache(b->tex.name);

   EXPECT(1,TOKEN_MISC,"(");


   /* Note: order of these two are not what you might expect */
   GETTOKEN(1,-1);
   ny=atoi(token);
   GETTOKEN(1,-1);
   nx=atoi(token);

   /* TODO : what are these numbers? flags, contents, value? */
   GETTOKEN(1,-1);
   GETTOKEN(1,-1);
   GETTOKEN(1,-1);

   EXPECT(1,TOKEN_MISC,")");


   b->x.q3c.sizex=nx;
   b->x.q3c.sizey=ny;
   b->x.q3c.a=Q_malloc(sizeof(vec2_t)*nx*ny);
   if (!b->x.q3c.a)
      return ERROR_NOMEM;
   
   b->num_verts=nx*ny;

   b->verts=Q_malloc(sizeof(vec3_t)*b->num_verts);
   if (!b->verts)
      return ERROR_NOMEM;

   b->num_faces=(nx-1)*(ny-1)/4;
   b->faces=Q_malloc(sizeof(face_t)*b->num_faces);
   if (!b->faces)
      return ERROR_NOMEM;

   memset(b->faces,0,sizeof(face_t)*b->num_faces);

   b->num_edges=0;
   b->edges=NULL;
   
      
   EXPECT(1,TOKEN_MISC,"(");

   for (i=0,k=0;i<ny;i++)
   {
      EXPECT(1,TOKEN_MISC,"(");
      for (j=0;j<nx;j++,k++)
      {
         EXPECT(0,TOKEN_MISC,"(");
         
         GETTOKEN(0,TOKEN_NUMBER);
         b->verts[k].x=atof(token);
         GETTOKEN(0,TOKEN_NUMBER);
         b->verts[k].y=atof(token);
         GETTOKEN(0,TOKEN_NUMBER);
         b->verts[k].z=atof(token);

         GETTOKEN(0,TOKEN_NUMBER);
         b->x.q3c.a[k].x=atof(token);
         GETTOKEN(0,TOKEN_NUMBER);
         b->x.q3c.a[k].y=atof(token);
         
         EXPECT(0,TOKEN_MISC,")");
      }
      EXPECT(1,TOKEN_MISC,")");
   }

   EXPECT(1,TOKEN_MISC,")");
   EXPECT(1,TOKEN_MISC,"}");
   EXPECT(1,TOKEN_MISC,"}");

   k=(nx-1)/2;
   for (i=0,f=b->faces;i<b->num_faces;i++,f++)
   {
      f->num_verts=9;
      f->verts=Q_malloc(sizeof(int)*f->num_verts);
      if (!f->verts)
         return ERROR_NOMEM;

      j=(i/k)*nx*2+2*(i%k);

      f->verts[0]=j;
      f->verts[1]=j+1;
      f->verts[2]=j+2;
      
      f->verts[3]=j+nx;
      f->verts[4]=j+nx+1;
      f->verts[5]=j+nx+2;
      
      f->verts[6]=j+2*nx;
      f->verts[7]=j+2*nx+1;
      f->verts[8]=j+2*nx+2;
   }

   RecalcNormals(b);
   CalcBrushCenter(b);

   b->Group=mapgroup;
   b->Entity=ent;

   return ERROR_NO;
}

static int LoadBrush_Old(map_uid_t ent,brush_t **res)
{
   cbrush_t B;
   cface_t *f;
   int i;
   brush_t *b;
   

   b=B_New(BRUSH_NORMAL);
   if (!b)
      return ERROR_NOMEM;
   *res=b;

   memset(&B,0,sizeof(cbrush_t));
   
   while (strcmp(token,"}"))
   {
      B.faces = Q_realloc(B.faces,sizeof(cface_t) * (B.num_faces+1));
      if (!B.faces)
         return ERROR_NOMEM;

      f=&B.faces[B.num_faces];
      memset(f,0,sizeof(cface_t));

      for (i=0;i<3;i++)
      {
         if (strcmp(token,"("))
            return ERROR_PARSE;

         GETTOKEN(0,TOKEN_NUMBER);
         f->planepts[i].x = atof(token);

         GETTOKEN(0,TOKEN_NUMBER);
         f->planepts[i].y = atof(token);

         GETTOKEN(0,TOKEN_NUMBER);
         f->planepts[i].z = atof(token);

         EXPECT(0,TOKEN_MISC,")");

         if (i!=2)
         {
            GETTOKEN(0,-1);
         }
      }

      if ((i=LoadTexInfo(&f->tex,&f->ba)))
         return i;

      B.num_faces++;

      GETTOKEN(1,-1);
   }

   CSG_CalcPlanes(&B);

   if (!CSG_BuildBrush(b,&B))
   {
      S_NewMessage("Warning: Invalid brush in .map file!");
      *res=NULL;
      return ERROR_NO;
   }

   b->Group=mapgroup;
   b->Entity=ent;

   return ERROR_NO;
}

static int LoadBrush_New(map_uid_t ent,brush_t **res)
{
   int i,j,k;
   brush_t *b;
   face_t *f;

   b=B_New(BRUSH_NORMAL);
   if (!b)
      return ERROR_NOMEM;
   *res=b;
   
   GETTOKEN(0,TOKEN_NUMBER);

   b->num_verts=atoi(token);

   b->verts=Q_malloc(sizeof(vec3_t)*b->num_verts);
   if (!b->verts)
      return ERROR_NOMEM;

   for (i=0;i<b->num_verts;i++)
   {
      GETTOKEN(1,TOKEN_NUMBER);
      b->verts[i].x=atof(token);

      GETTOKEN(0,TOKEN_NUMBER);
      b->verts[i].y=atof(token);

      GETTOKEN(0,TOKEN_NUMBER);
      b->verts[i].z=atof(token);
   }

   b->num_faces=0;
   b->faces=NULL;

   GETTOKEN(1,-1);

   while (strcmp(token,"}"))
   {
      b->faces=Q_realloc(b->faces,sizeof(face_t)*(b->num_faces+1));
      if (!b->faces)
         return ERROR_NOMEM;

      f=&b->faces[b->num_faces++];
      memset(f,0,sizeof(face_t));

      f->num_verts=atoi(token);

      f->verts=Q_malloc(sizeof(int)*f->num_verts);
      if (!f->verts)
         return ERROR_NOMEM;

      EXPECT(0,TOKEN_MISC,"(");

      for (i=0;i<f->num_verts;i++)
      {
         GETTOKEN(0,TOKEN_NUMBER);
         f->verts[i]=atoi(token);
      }

      EXPECT(0,TOKEN_MISC,")");

      if ((i=LoadTexInfo(&f->tex,&f->a)))
         return i;

      GETTOKEN(1,-1);
   }

   b->num_edges=0;
   b->edges=NULL;

   for (i=0;i<b->num_faces;i++)
   {
      int v1,v2;

      f=&b->faces[i];

      for (j=0;j<f->num_verts;j++)
      {
         v1=f->verts[j];
         if (j==f->num_verts-1)
            v2=f->verts[0];
         else
            v2=f->verts[j+1];

         if (v2<v1)
         {
            k=v2;
            v2=v1;
            v1=k;
         }

         for (k=0;k<b->num_edges;k++)
         {
            if ((b->edges[k].startvertex==v1) &&
                (b->edges[k].endvertex==v2))
               break;
         }
         if (k==b->num_edges)
         {
            b->edges=Q_realloc(b->edges,sizeof(edge_t)*(b->num_edges+1));
            if (!b->edges)
               return ERROR_NOMEM;
            b->edges[b->num_edges].startvertex=v1;
            b->edges[b->num_edges].endvertex=v2;
            b->num_edges++;
         }
      }
   }

   b->Group=mapgroup;
   b->Entity=ent;

   return ERROR_NO;
}


static int LoadBrush(map_uid_t ent,brush_t **res)
{
   mapgroup=M.WorldGroup;

   GETTOKEN(1,TOKEN_NAME|TOKEN_MISC);
   if (!strcmp(token,"patchDef2"))
   /* Quake 3 style bezier patch 'brush' */
      return LoadBrush_Q3C(ent,res);
   else
   if (!strcmp(token,"("))
   /* Old-style .map brush */
      return LoadBrush_Old(ent,res);
   else
   if (!strcmp(token,":"))
   /* New style .map brush */
      return LoadBrush_New(ent,res);
   else
      return ERROR_PARSE;
}


static int LoadEntity(entity_t **res,brush_t **bfirst,brush_t **blast,int *num
#if TODO
,int is_group
#endif
   )
{
   entity_t *e;
   int i;
   brush_t *first,*last;
   brush_t *b;

   mapgroup=M.WorldGroup;

   if (strcmp(token,"{"))
      return ERROR_PARSE;

   GETTOKEN(1,-1);

#if TODO
   if (strcmp(token,"*") || !is_group)
#endif
   {
      e=E_New();
      if (!e)
         return ERROR_NOMEM;
      e->uid=GetEntityUID();
      e->real=1;
   
      *res=e;
   
      e->Group=mapgroup;
   
      while (strcmp(token,"}") && strcmp(token,"{"))
      {
         strcpy(buf,&token[1]);
         buf[strlen(buf)-1]=0;
   
         GETTOKEN(0,TOKEN_STRING);
   
         strcpy(token,&token[1]);
         token[strlen(token)-1]=0;
   
         SetKeyValue(e,buf,token);
   
         GETTOKEN(1,-1);
      }
   }
#if TODO
   else
   {
      e=M.WorldSpawn;
      *res=NULL;

      GETTOKEN(1,-1);
   }
#endif

   first=last=NULL;
   *num=0;

   while (!strcmp(token,"{"))
   {
      i=LoadBrush(e->uid,&b);
      if (i)
         return i;

      if (b)
      {
         b->uid=GetBrushUID();
         b->real=1;

         if (!last)
            last=b;

         if (first)
            first->Last=b;
         b->Next=first;
         b->Last=NULL;
         first=b;

         (*num)++;
      }

      GETTOKEN(1,-1);
   }

   if (strcmp(token,"}"))
      return ERROR_PARSE;

   *bfirst=first;
   *blast=last;

   return ERROR_NO;
}


int QMap_Load(const char *filename)
{
   int i;
   entity_t *e;
   brush_t *b1,*b2;
   int n_brush;

   Grp_CreateWorldGroup();

   if (!TokenFile(filename,TOKEN_C|TOKEN_NUMBER|TOKEN_STRING|TOKEN_MISC,Comments)) /* TODO: comment parsing */
   {
      S_Error("QMap_LoadMap","Unable to load '%s'!",filename);
      return 0;
   }

   infogroup=0;
   infogroup_p=NULL;

#if TODO
   num_extra=0;
   extra_tex=NULL;

   load_camera=0;
#endif

   M.num_entities = 0;
   M.num_brushes  = 0;
   M.BrushHead    = NULL;
   M.EntityHead   = NULL;
   M.WorldSpawn   = 0;

   InitViewports();

   while (TokenGet(1,-1))
   {
      i=LoadEntity(&e,&b1,&b2,&n_brush/*TODO,0*/);
      if (i)
      {
         S_Error("QMap_LoadMap","'%s', line %i: %s",
            filename,token_linenum,errors[i]);
         return 0;
      }

      if (b1)
      {
         if (M.BrushHead) M.BrushHead->Last=b2;
         b2->Next=M.BrushHead;
         M.BrushHead=b1;
         M.num_brushes+=n_brush;
      }

      if (M.EntityHead) M.EntityHead->Last=e;
      e->Next = M.EntityHead;
      e->Last = NULL;
      M.EntityHead = e;
      M.num_entities++;

      if (!M.WorldSpawn)  // assume the first entity is the worldspawn entity
         M.WorldSpawn = e->uid;
   }

   TokenDone();

#if TODO
   if (!load_camera)
#endif
      InitViewports();

#if TODO
   ReadCache(1);

   if (Game.tex.cache)
   {
      for (i=0;i<num_extra;i++)
         ReadMIPTex(extra_tex[i],0);

      Q_free(extra_tex);
      num_extra=0;
   }
#endif
   Map_Update();

   return 1;
}


#if TODO
int QMap_LoadGroup(const char *filename)
{
   int i;
   entity_t *e;
   brush_t *b1,*b2;
   int n_brush;

/*   int dx,dy,dz;
   int x,y,z;
   viewport_t *vp;
   char *value;
   brush_t *b;*/


   if (!TokenFile(filename,T_C|TOKEN_NUMBER|T_STRING|TOKEN_MISC,Comments))
   {
      HandleError("QMap_LoadGroup","Unable to load '%s'!",filename);
      return 0;
   }

/* This code assumes the saved group is centered at origin, which
   isn't true at the moment. Leave the group where it is.

   Move(M.display.active_vport,MOVE_FORWARD,&dx,&dy,&dz,128);
   vp=&M.display.vport[M.display.active_vport];

   dx+=SnapPointToGrid(vp->camera_pos.x);
   dy+=SnapPointToGrid(vp->camera_pos.y);
   dz+=SnapPointToGrid(vp->camera_pos.z);*/

   while (TokenGet(1,-1))
   {
      i=LoadEntity(&e,&b1,&b2,&n_brush,1);
      if (i)
      {
         HandleError("QMap_LoadMap","'%s', line %i: %s",
            filename,token_linenum,errors[i]);
         return 0;
      }

/*      if (e)
      {
         value=GetKeyValue(e,"origin");
         if (value)
         {
            sscanf(value,"%i %i %i", &x, &y, &z);
            sprintf(value,"%i %i %i",x+dx,y+dy,z+dz);
            SetKeyValue(e,"origin",value);
         }
      }

      for (b=b1;b;b=b->Next)
      {
         for (i=0;i<b->num_verts;i++)
         {
            b->verts[i].x += dx;
            b->verts[i].y += dy;
            b->verts[i].z += dz;
         }
      }*/

      if (b1)
      {
         if (M.BrushHead) M.BrushHead->Last=b2;
         b2->Next=M.BrushHead;
         M.BrushHead=b1;
         M.num_brushes+=n_brush;
      }

      if (e)
      {
         if (M.EntityHead) M.EntityHead->Last=e;
         e->Next = M.EntityHead;
         e->Last = NULL;
         M.EntityHead = e;
         M.num_entities++;
      }
   }

   TokenDone();

   RecalcAllNormals();

   return 1;
}
#endif


/* Map saving stuff. */

static void SaveCurve(FILE *fp,brush_t *b)
{
   int nx,ny,i,j;

   if (!b->num_faces)
      return;

   nx=b->faces[0].verts[3];
   ny=(b->num_faces/((nx-1)/2))*2+1;

   if ( (b->num_faces%((nx-1)/2)) || (b->num_faces%((ny-1)/2)) ||
        (b->num_verts!=nx*ny) ||
        (nx!=b->x.q3c.sizex) || (ny!=b->x.q3c.sizey)
      )
   {
      S_Error("QMap_SaveCurve",
         "Curve too complex to save in Quake 3 .map format!");
      return;
   }

   fprintf(fp,"{\n");
   if (b->Group!=M.WorldGroup)
   {
      group_t *g=UID_Group(b->Group);
      if (g)
         fprintf(fp,"//Quest Brush Group: %s\n",g->fullname);
   }

   fprintf(fp,"patchDef2\n{\n%s\n( %i %i 0 0 0 )\n(\n",
      b->tex.name[0]?b->tex.name:"no_texture",
      ny,nx);

   for (i=0;i<ny;i++)
   {
      fprintf(fp,"( ");
      for (j=0;j<nx;j++)
      {
         fprintf(fp,"( %g %g %g %g %g ) ",
            snaptoint(b->verts[i*nx+j].x),
            snaptoint(b->verts[i*nx+j].y),
            snaptoint(b->verts[i*nx+j].z),

            snaptoint(b->x.q3c.a[i*nx+j].x),
            snaptoint(b->x.q3c.a[i*nx+j].y)
            );
      }
      fprintf(fp,")\n");
   }

   fprintf(fp,")\n}\n}\n");
}

static void SaveBrush(FILE *fp,brush_t *b)
{
   int i,j;
   face_t *f;
   cface_t F;

   switch (b->bt->type)
   {
   case BRUSH_NORMAL:
      break;

   case BRUSH_Q3_CURVE:
      SaveCurve(fp,b);
      return;

   default:
      return;
   }

   fprintf(fp, "{\n");
   if (b->Group!=M.WorldGroup)
   {
      group_t *g=UID_Group(b->Group);
      if (g)
         fprintf(fp,"//Quest Brush Group: %s\n",g->fullname);
   }

   if (cv_qmap_map_format->i==2)
   {
      fprintf(fp,":%i\n",b->num_verts);
      for (i=0;i<b->num_verts;i++)
         fprintf(fp,"%g %g %g\n",
            snaptoint(b->verts[i].x),
            snaptoint(b->verts[i].y),
            snaptoint(b->verts[i].z));
   }

   for (i=0; i<b->num_faces; i++)
   {
      f=&b->faces[i];

      if (cv_qmap_map_format->i==2)
      {
         fprintf(fp,"%i ( ",f->num_verts);

         for (j=0;j<f->num_verts;j++)
            fprintf(fp,"%i ",f->verts[j]);

         fprintf(fp,") ");
      }
      else /* TODO: check if our three vertices are on a line, pick others til
              we get a valid plane (if possible) */
      if (cv_qmap_map_format->i==1)
      {
            fprintf(fp, "( %g %g %g ) ( %g %g %g ) ( %g %g %g ) ",
                    b->verts[f->verts[0]].x,
                    b->verts[f->verts[0]].y,
                    b->verts[f->verts[0]].z,
                    b->verts[f->verts[1]].x,
                    b->verts[f->verts[1]].y,
                    b->verts[f->verts[1]].z,
                    b->verts[f->verts[2]].x,
                    b->verts[f->verts[2]].y,
                    b->verts[f->verts[2]].z);
      }
      else
      {  /* old brush save */
         /* See if we should use the actual vertex locations (if they
            are snapped to integers, they're preferable to use), or
            if we should generate a large box plane and use it. */
         for (j=0;j<f->num_verts;j++)
         {
            if (fabs(b->verts[f->verts[j]].x - floor(b->verts[f->verts[j]].x + .5)) > 0.01)
               break;
            if (fabs(b->verts[f->verts[j]].y - floor(b->verts[f->verts[j]].y + .5)) > 0.01)
               break;
            if (fabs(b->verts[f->verts[j]].z - floor(b->verts[f->verts[j]].z + .5)) > 0.01)
               break;
         }

         if (j==f->num_verts)
         {
            /* Using actual vertices */
            fprintf(fp, "( %i %i %i ) ( %i %i %i ) ( %i %i %i ) ",
                    realtoint(b->verts[f->verts[0]].x),
                    realtoint(b->verts[f->verts[0]].y),
                    realtoint(b->verts[f->verts[0]].z),
                    realtoint(b->verts[f->verts[1]].x),
                    realtoint(b->verts[f->verts[1]].y),
                    realtoint(b->verts[f->verts[1]].z),
                    realtoint(b->verts[f->verts[2]].x),
                    realtoint(b->verts[f->verts[2]].y),
                    realtoint(b->verts[f->verts[2]].z));
         }
         else
         {
            /* Using MakeBoxOnPlane for save */
            F.p.n.x = f->normal.x;
            F.p.n.y = f->normal.y;
            F.p.n.z = f->normal.z;
            F.p.dist = f->dist;

            CSG_MakeBoxOnPlane(&F);

            fprintf(fp, "( %i %i %i ) ( %i %i %i ) ( %i %i %i ) ",
                     realtoint(F.pts[0].x),
                     realtoint(F.pts[0].y),
                     realtoint(F.pts[0].z),
                     realtoint(F.pts[1].x),
                     realtoint(F.pts[1].y),
                     realtoint(F.pts[1].z),
                     realtoint(F.pts[2].x),
                     realtoint(F.pts[2].y),
                     realtoint(F.pts[2].z));
         }
      } /* end old plane code */

      SaveTexInfo(f,fp);

      fprintf(fp,"\n");
   }
   fprintf(fp, "}\n");
}


static void WriteHeader(FILE *fp)
{
#if TODO
   viewport_t *vp;
   int i;
#endif

   fprintf(fp, "// --\n");
   fprintf(fp, "// -- Created with Quest\n");
   fprintf(fp, "// --\n");

#if TODO
   fprintf(fp, "// ** %i num_vports\n",M.display.num_vports);

   for (i=0; i<M.display.num_vports; i++)
   {
      vp=&M.display.vport[i];

      fprintf(fp, "// ** %i spos %.5f %.5f %.5f %.5f\n",i,
         vp->f_xmin,vp->f_ymin,vp->f_xmax,vp->f_ymax);

      fprintf(fp, "// ** %i pos %i %i %i\n",i,
         vp->camera_pos.x,vp->camera_pos.y,vp->camera_pos.z);
      fprintf(fp, "// ** %i dir %i\n",i,vp->camera_dir);
      fprintf(fp, "// ** %i fullbright %i\n",i,vp->fullbright);
      fprintf(fp, "// ** %i zoom_amt %2.4f\n",i,vp->zoom_amt);
      fprintf(fp, "// ** %i mode %i\n",i,vp->mode);
      if (!vp->axis_aligned)
         fprintf(fp,"// ** %i angle %i %i %i\n",
            i,vp->rot_x,vp->rot_y,vp->rot_z);
   }
#endif
}


static void WriteGroupInfo(FILE *fp,group_t *g)
{
   fprintf(fp,"// Quest Group Name: %s\n",g->fullname);
   fprintf(fp,"// Quest Group Flags: %i\n",g->flags);
   fprintf(fp,"// Quest Group Color: %i\n",(unsigned char)g->color);
}

static void WriteGroups_r(FILE *fp,group_t *g)
{
   map_uid_t uc;
   group_t *c;
   WriteGroupInfo(fp,g);
   if (g->child)
   {
      for (uc=g->child;;uc=c->cnext)
      {
         c=UID_Group(uc);
         if (!c->cnext)
            break;
      }
      for (;uc;uc=c->clast)
      {
         c=UID_Group(uc);
         WriteGroups_r(fp,c);
      }
   }
}

static void WriteGroups(FILE *fp)
{
   group_t *g;

/* We need to write parents before children so loading will work */
   g=UID_Group(M.WorldGroup);
   if (g) WriteGroups_r(fp,g);
}


static void WriteEntity(FILE *fp,entity_t *e)
{
   int i;

   fprintf(fp,"{\n");
   if (e->Group!=M.WorldGroup)
   {
      group_t *g=UID_Group(e->Group);
      if (g)
         fprintf(fp,"//Quest Entity Group: %s\n", g->fullname);
   }

   for (i=0; i<e->num_keys; i++)
   {
      fprintf(fp,"\"%s\" \"%s\"\n", e->key[i], e->value[i]);
   }
}


int QMap_Save(const char *filename)
{
#if TODO
   int  i;
#endif

   FILE *fp;

   entity_t *e;

   brush_t *LastBrush;
   brush_t *b;

   if (!filename[0])
      return 0;

   fp=fopen(filename,"wt");
   if (!fp)
   {
      S_Error("QMap_SaveMap", "Unable to create file '%s'!",filename);
      return 0;
   }

   WriteHeader(fp);

   WriteGroups(fp);

#if TODO
   if (Game.tex.cache)
   {
      for (i=0;i<M.num_textures;i++)
      {
         fprintf(fp,"// T: %s\n",M.Cache[i].name);
      }
   }
#endif

   /* Locate the last entity (will be worldspawn since our lists were built
      backwards) */
   if (M.EntityHead)
      for (e=M.EntityHead; e->Next; e=e->Next) { }
   else
      e=NULL;
   if (M.BrushHead)
   {
      for (b=M.BrushHead; b->Next; b=b->Next) { }
      LastBrush=b;
   }
   else
      LastBrush=NULL;

   for (; e; e=e->Last)
   {
      WriteEntity(fp,e);

      for (b=LastBrush; b; b=b->Last)
         if (b->Entity == e->uid)
            SaveBrush(fp,b);

      fprintf(fp, "}\n");
   }

   fclose(fp);

   M.modified=0;

   return 1;
}


#if TODO
int QMap_SaveGroup(const char *filename,group_t *g)
{
   FILE *fp;

   brush_t  *b;
   entity_t *e;

   int found;


   fp=fopen(filename,"wt");
   if (!fp)
   {
      HandleError("QMap_LoadPrefabBrush", "Unable to create '%s'!",filename);
      return 0;
   }

   WriteGroupInfo(fp,g);

   /* Write entities */
   for (e=M.EntityHead; e; e=e->Next)
   {
      found=0;
      if ((e->Group==g) && (e!=M.WorldSpawn))
      {
         WriteEntity(fp,e);
         found=1;
      }

      /* Write brushes w/entity, if any exist */
      for (b=M.BrushHead;b;b=b->Next)
      {
         if ((b->Group == g) && (b->EntityRef == e))
         {
            /* If it's the first brush in the entity, write out all
               the key/values and brackets */
            if (!found)
            {
               found = 1;
               if (e == M.WorldSpawn)
               {
                  fprintf(fp, "{*\n");
               }
               else
               {
                  WriteEntity(fp,e);
               }
            }

            SaveBrush(fp,b);
         }
      }
      if (found)
         fprintf(fp,"}\n");
   }
   fclose(fp);

   return 1;
}


int QMap_SaveVisible(const char *filename,int add_shell)
{
   int  found_start;
   FILE *fp;
   entity_t  *e;
   brush_t   *LastBrush;
   brush_t   *b;

   vec3_t min,max;

   int numb;


   if (add_shell)
   {
/*
 Find min and max values to create a box around the level so vis
will still work.
*/
      vec3_t temp;
      int i;

      min.x=min.y=min.z= 10000;
      max.x=max.y=max.z=-10000;
#define ChkVec(v,a) \
   if (v.a<min.a) min.a=v.a; \
   if (v.a>max.a) max.a=v.a;
      for (b=M.BrushHead;b;b=b->Next)
      {
         if (b->Group->flags & 0x02)
            continue;

         for (i=0;i<b->num_verts;i++)
         {
            ChkVec(b->verts[i],x);
            ChkVec(b->verts[i],y);
            ChkVec(b->verts[i],z);
         }
      }

      for (e=M.EntityHead;e;e=e->Next)
      {
         if (e->Group->flags & 0x02)
            continue;

         if (!GetKeyValue(e,"origin"))
            continue;

         sscanf(GetKeyValue(e,"origin"),"%f %f %f",&temp.x,&temp.y,&temp.z);
         ChkVec(temp,x);
         ChkVec(temp,y);
         ChkVec(temp,z);
      }

#undef ChkVec
/*      min.x-=32;
      min.y-=32;
      min.z-=32;
      max.x+=32;
      max.y+=32;
      max.z+=32;*/
   }

   found_start = FALSE;

   fp=fopen(filename,"wt");
   if (!fp)
   {
      HandleError("QMap_CreateVisibleMap", "Unable to create '%s'!",filename);
      return 0;
   }

   WriteHeader(fp);
   WriteGroups(fp);

   /* Locate the last entity (will be WorldSpawn since our lists were built
      backwards) */
   for (e=M.EntityHead; e->Next; e=e->Next) { }
   if (M.BrushHead)
   {
      for (b=M.BrushHead; b->Next; b=b->Next) { }
      LastBrush=b;
   }
   else
      LastBrush=NULL;

   for (; e; e=e->Last)
   {
      if (e!=M.WorldSpawn)
      {
         if (e->Group->flags & 0x02)
         {
            continue;
         }

         if (!GetKeyValue(e,"origin"))
         {
            for (b=LastBrush; b; b=b->Last)
            {
               if (b->Group->flags & 0x02)
                  continue;
               if (b->EntityRef == e)
                  break;
            }
            if (!b)
            {
               continue;
            }
         }
      }

      WriteEntity(fp,e);

      if (GetKeyValue(e,"classname"))
         if (!strcmp(GetKeyValue(e,"classname"),"info_player_start"))
            found_start=1;

      if (GetKeyValue(e,"classname"))
      {
         if (add_shell && !strcmp(GetKeyValue(e,"classname"),"worldspawn"))
         {
#define AddPlane(p00,p01,p02,p10,p11,p12,p20,p21,p22) \
   fprintf(fp,"( %0.0f %0.0f %0.0f ) " \
              "( %0.0f %0.0f %0.0f ) " \
              "( %0.0f %0.0f %0.0f ) " \
              "region 0 0 0 1 1\n", \
      p00,p01,p02,p10,p11,p12,p20,p21,p22)
         
#define AddBrush(x1,y1,z1,x2,y2,z2) \
   fprintf(fp,"{\n"); \
   \
   AddPlane(x1,y1,z1,x2,y1,z1,x1,y2,z1); \
   AddPlane(x2,y2,z2,x2,y1,z2,x1,y2,z2); \
   \
   AddPlane(x1,y1,z1,x1,y2,z1,x1,y1,z2); \
   AddPlane(x2,y2,z2,x2,y2,z1,x2,y1,z2); \
   \
   AddPlane(x1,y1,z1,x1,y1,z2,x2,y1,z1); \
   AddPlane(x2,y2,z2,x1,y2,z2,x2,y2,z1); \
   \
   fprintf(fp,"}\n");
         
            AddBrush(min.x-16,min.y,min.z,min.x   ,max.y,max.z);
            AddBrush(max.x   ,min.y,min.z,max.x+16,max.y,max.z);

            AddBrush(min.x,min.y-16,min.z,max.x,min.y   ,max.z);
            AddBrush(min.x,max.y   ,min.z,max.x,max.y+16,max.z);

            AddBrush(min.x,min.y,min.z-16,max.x,max.y,min.z   );
            AddBrush(min.x,min.y,max.z   ,max.x,max.y,max.z+16);
#undef AddBrush
#undef AddPlane
         }
      }

      numb=0;
      for (b=LastBrush; b; b=b->Last)
      {
         if (b->Group->flags & 0x02)
            continue;
         if (b->EntityRef == e)
         {
            SaveBrush(fp,b);
         }
      }
      fprintf(fp, "}\n");
   }

   if (!found_start)
   {
      viewport_t *vp;

      vp=&M.display.vport[M.display.active_vport];

      fprintf(fp,"{\n");
      fprintf(fp,"\"classname\" \"info_player_start\"\n");
      fprintf(fp,"\"origin\" \"%i %i %i\"\n",
              vp->camera_pos.x,
              vp->camera_pos.y,
              vp->camera_pos.z);
      fprintf(fp,"}\n");
   }

   fclose(fp);

   return 1;
}
#endif


void QMap_Init(void)
{
static int did_init=0;
   if (did_init)
      return;
   did_init=1;

   cv_qmap_map_format=
      Cfg_DefVar("qmap.map_format","2",
         "Which map format the Quake engine map saver should use.\n"
         "\n"
         "0: Original brush-style format, integer coordinates\n"
         "1: Original brush-style format, real coordinates\n"
         "2: Quest's new brush-style format\n"
         "\n"
         "Format #2 removes vertex drifting errors, produces smaller files, "
         "and it loads quicker. Format #1 reduces vertex drifting errors. "
         "However, not all compiling utilities support format "
         "1 or 2.");
}

