/*
entity.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "entity.h"

#include "entclass.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"
#include "s_gui.h"
#include "s_render.h"
#include "uid.h"
#include "undo.h"


entity_t *E_New(void)
{
   entity_t *e;

   e=Q_malloc(sizeof(entity_t));
   if (!e) return NULL;
   memset(e,0,sizeof(entity_t));

   return e;
}

void E_FreeC(entity_t *e)
{
   int i;

   QASSERT(!e->real || (!e->Next && !e->Last));

   S_R_EntityFree(e);
   
   for (i=0;i<e->num_keys;i++)
   {
      if (e->key[i]) Q_free(e->key[i]);
      if (e->value[i]) Q_free(e->value[i]);
   }
   Q_free(e->key);
   Q_free(e->value);
   e->key=NULL;
   e->value=NULL;
   e->num_keys=0;
}

void E_Free(entity_t *e)
{
   E_FreeC(e);
   Q_free(e);
}

void E_Link(entity_t *e)
{
   QASSERT(!e->Next && !e->Last);

   if (M.EntityHead!=NULL) M.EntityHead->Last = e;
   e->Next = M.EntityHead;
   e->Last = NULL;
   M.EntityHead = e;
   e->real=1;
   e->VNext=e->VLast=NULL;

   e->uid=GetEntityUID();

   M.num_entities++;

   U_EntityAdd(e);
}

void E_Unlink(entity_t *e)
{
   QASSERT(e->Next || e->Last || e==M.EntityHead);
   QASSERT(e->real);

   U_EntityDelete(e);

   M.num_entities--;
   
   if (e->Next != NULL) e->Next->Last = e->Last;
   if (e->Last != NULL) e->Last->Next = e->Next;
   if (e==M.EntityHead) M.EntityHead=e->Next;

   e->Next=NULL;
   e->Last=NULL;
   e->VNext=e->VLast=NULL;
   e->real=0;
}

entity_t *E_Duplicate(entity_t *dest,const entity_t *source)
{
   int i;
   if (!dest)
   {
      dest=Q_malloc(sizeof(entity_t));
      if (!dest) return NULL;
   }
   memset(dest,0,sizeof(entity_t));
   dest->num_keys=source->num_keys;
   
   dest->key=Q_malloc(sizeof(char *)*dest->num_keys);
   dest->value=Q_malloc(sizeof(char *)*dest->num_keys);
   if (!dest->key || !dest->value) return NULL;

   for (i=0;i<dest->num_keys;i++)
   {
      dest->key[i]=Q_strdup(source->key[i]);
      dest->value[i]=Q_strdup(source->value[i]);
      if (!dest->key[i] || !dest->value[i]) return NULL;
   }
   
   return dest;
}

int E_Size(const entity_t *e)
{
   int s,i;

   for (s=0,i=0;i<e->num_keys;i++)
   {
      s+=strlen(e->key[i])+1;
      s+=strlen(e->value[i])+1;
   }
   s+=sizeof(char *)*e->num_keys*2;
   return s;
}

void E_Update(entity_t *e)
{
   const char *c=GetKeyValue(e,"classname");
   if (!c || !(e->eclass=FindClass(c)))
      e->eclass=&eclass_dummy;
      
   S_R_EntityUpdate(e);
}


const char *GetKeyValue(const entity_t *ent,const char *key)
{
   int   i;
   for (i=0; i<ent->num_keys; i++)
      if (!strcmp(ent->key[i], key))
         return ent->value[i];
   return NULL;
}

static void CreateKeyValue(entity_t *ent,const char *key,const char *value)
{
   char **c;
   
   U_EntityModify(ent);
   
   c=(char **)Q_realloc(ent->key,sizeof(char *)*(ent->num_keys+1));
   if (!c)
   {
      S_Error("CreateKeyValue","Out of memory!");
      return;
   }
   ent->key=c;
   c=(char **)Q_realloc(ent->value,sizeof(char *)*(ent->num_keys+1));
   if (!c)
   {
      S_Error("CreateKeyValue","Out of memory!");
      return;
   }
   ent->value=c;

   ent->key[ent->num_keys] = Q_strdup(key);
   ent->value[ent->num_keys] = Q_strdup(value);
   if (!ent->key[ent->num_keys] || !ent->key[ent->num_keys])
   {
      S_Error("CreateKeyValue","Out of memory!");
      return;
   }

   ent->num_keys++;
}

void SetKeyValue(entity_t *ent,const char *key,const char *value)
{
   int   i;
   char *v;

   /* If setting to null, remove key */
   if (!value)
   {
      RemoveKeyValue(ent,key);
      return;
   }

   for (i=0;i<ent->num_keys;i++)
   {
      if (!strcmp(ent->key[i],key))
      {
         U_EntityModify(ent);
         v=Q_realloc(ent->value[i],strlen(value)+1);
         if (!v)
         {
            S_Error("SetKeyValue","Out of memory!");
            return;
         }
         ent->value[i]=v;
         strcpy(ent->value[i],value);
         return;
      }
   }
   CreateKeyValue(ent,key,value);
}

void RemoveKeyValue(entity_t *ent,const char *key)
{
   int i;
   char **c;

   if (!ent->num_keys)
      return;
   for (i=0; i<ent->num_keys; i++)
      if (strcmp(ent->key[i], key) == 0)
         break;
   if (i==ent->num_keys)
      return;
   U_EntityModify(ent);
   
   Q_free(ent->key[i]);
   Q_free(ent->value[i]);
   for (; i<ent->num_keys-1; i++)
   {
      ent->key[i]=ent->key[i+1];
      ent->value[i]=ent->value[i+1];
   }

   ent->num_keys--;
   if (ent->num_keys)
   { /* These shouldn't fail, but if they do, we can just ignore it */
      if ((c=(char **)Q_realloc(ent->key,sizeof(char *)*(ent->num_keys-1))))
         ent->key=c;
      if ((c=(char **)Q_realloc(ent->value,sizeof(char *)*(ent->num_keys-1))))
         ent->value=c;
   }
   else
   {
      Q_free(ent->key);
      Q_free(ent->value);
      ent->key=ent->value=NULL;
   }
}


int Target_GetStrNum(const char *v)
{
   char *bp;
   int i;

   if (v[0]!='t')
      return -1;
   i=strtol(&v[1],&bp,0);
   if (*bp)
      return -1;
   return i;
}

int Target_GetNum(entity_t *e,const char *key)
{
   const char *v;

   v=GetKeyValue(e,key);
   if (!v)
      return -1;
   return Target_GetStrNum(v);
}

int Target_GetLastNum(const char *key)
{
   int last;
   entity_t *e;
   int num;

   last=-1;
   for (e=M.EntityHead;e;e=e->Next)
   {
      num=Target_GetNum(e,key);
      if (num>last) last=num;
   }

   return last;
}


void E_Edit(int nents,entity_t **ents)
{
	int i,j;
   entity_t *Ent;
   entity_t *a;

   Ent=Q_malloc(sizeof(entity_t));
   if (!Ent)
   {
      S_Error("E_Edit","Out of memory!");
      return;
   }
   memset(Ent,0,sizeof(entity_t));
   Ent->num_keys=0;

   for (i=0;i<nents;i++)
   {
/*      printf("entity %i:\n",i);
      DumpEntity(ents[i]);*/

      a=ents[i];
      for (j=0;j<a->num_keys;j++)
      {
         if (!GetKeyValue(Ent,a->key[j]))
         {
            if (i)
               SetKeyValue(Ent,a->key[j],"?");
            else
               SetKeyValue(Ent,a->key[j],a->value[j]);
         }
         else
         {
            if (strcmp(a->value[j],GetKeyValue(Ent,a->key[j])))
            {
               SetKeyValue(Ent,a->key[j],"?");
            }
         }
      }
      for (j=0;j<Ent->num_keys;j++)
      {
         if (!GetKeyValue(a,Ent->key[j]))
            SetKeyValue(Ent,Ent->key[j],"?");
      }
   }
/*   printf("total: %i\n",nents);
   printf("Combined:\n");
   DumpEntity(Ent);*/

   if (S_GUI_EditEntity(Ent))
   {
      for (i=0;i<nents;i++)
         U_EntityModify(ents[i]);

   /*   printf("After:\n");
      DumpEntity(Ent);*/
   
      for (i=0;i<Ent->num_keys;i++)
      {
         if (strcmp(Ent->value[i],"?"))
         {
            for (j=0;j<nents;j++)
            {
               SetKeyValue(ents[j],Ent->key[i],Ent->value[i]);
            }
         }
      }
      for (i=0;i<nents;i++)
      {
         a=ents[i];
   
         for (j=0;j<a->num_keys;j++)
         {
            if (!GetKeyValue(Ent,a->key[j]))
            {
               RemoveKeyValue(a,a->key[j]);
               j--;
            }
         }
      }
/*      for (i=0;i<nents;i++)
      {
         printf("entity %i:\n",i);
         DumpEntity(ents[i]);
      }*/
   }

   for (i=0;i<Ent->num_keys;i++)
   {
      Q_free(Ent->key[i]);
      Q_free(Ent->value[i]);
   }
   Q_free(Ent->key);
   Q_free(Ent->value);
   Q_free(Ent);
}

