#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "map.h"

#include "brush.h"
#include "camera.h"
#include "config.h"
#include "entclass.h"
#include "game.h"
#include "geom.h"
#include "hook.h"
#include "quest.h"
#include "s_texture.h"
#include "status.h"


/* TODO: better names */
scalar_t SnapPointToGrid(scalar_t pt)
{
	if (pt>0)
	{
		return floor(pt / cv_edit_snap_size->i + 0.5) * cv_edit_snap_size->i;
	}
	if (pt<0)
	{
		return  ceil(pt / cv_edit_snap_size->i - 0.5) * cv_edit_snap_size->i;
	}
	return pt;
}

void SnapVecToGrid(vec3_t *v)
{
   v->x=SnapPointToGrid(v->x);
   v->y=SnapPointToGrid(v->y);
   v->z=SnapPointToGrid(v->z);
}


scalar_t SnapPointToInt(scalar_t p)
{
   int pi=floor(p+0.5);
   if (fabs(p-pi)<0.001) return pi;
   return p;
}

void SnapVecToInt(vec3_t *v)
{
   v->x=SnapPointToInt(v->x);
   v->y=SnapPointToInt(v->y);
   v->z=SnapPointToInt(v->z);
}


/* TODO: global DELTA and EPSILON defines */
int VertsAreEqual(vec3_t one,vec3_t two)
{
   if ((fabs(one.x - two.x) < 0.1) &&
       (fabs(one.y - two.y) < 0.1) &&
       (fabs(one.z - two.z) < 0.1))
      return 1;
   else
      return 0;
}

int EdgesAreEqual(edge_t one,edge_t two)
{
   if ((one.startvertex == two.startvertex) &&
       (one.endvertex   == two.endvertex))
      return 1;

   if ((one.startvertex == two.endvertex) &&
       (one.endvertex   == two.startvertex))
      return 1;

   return 0;
}


void CalcBrushCenter(brush_t *b)
{
   int i, j;

   b->center.x = 0;
   b->center.y = 0;
   b->center.z = 0;

   for (i=0; i<b->num_verts; i++)
   {
      b->center.x += b->verts[i].x;
      b->center.y += b->verts[i].y;
      b->center.z += b->verts[i].z;
   }
   b->center.x /= b->num_verts;
   b->center.y /= b->num_verts;
   b->center.z /= b->num_verts;

   for (i=0; i<b->num_faces; i++)
   {
      b->faces[i].center.x = 0;
      b->faces[i].center.y = 0;
      b->faces[i].center.z = 0;
      for (j=0; j<b->faces[i].num_verts; j++)
      {
         b->faces[i].center.x += b->verts[b->faces[i].verts[j]].x;
         b->faces[i].center.y += b->verts[b->faces[i].verts[j]].y;
         b->faces[i].center.z += b->verts[b->faces[i].verts[j]].z;
      }
      b->faces[i].center.x /= b->faces[i].num_verts;
      b->faces[i].center.y /= b->faces[i].num_verts;
      b->faces[i].center.z /= b->faces[i].num_verts;
   }
}


void RecalcNormals(brush_t *b)
{
	int      j, k;
	int      k_next, k_prev, k_cur;
	face_t  *f;
	vec3_t   vec1, vec2, norm;
	vec3_t  *vertarray;

   f=b->faces;
   vertarray = b->verts;
	for (j=0; j<b->num_faces; j++,f++)
	{

		norm.x = norm.y = norm.z = 0;
		k = 1;
		while ((norm.x == 0) &&
		       (norm.y == 0) &&
		       (norm.z == 0) &&
		       (k <= f->num_verts))
      {

			k_cur = k % f->num_verts;
			k_next = (k + 1) % f->num_verts;
			k_prev = (k + f->num_verts - 1) % f->num_verts;

			vec1.x = vertarray[f->verts[k_cur]].x - vertarray[f->verts[k_prev]].x;
			vec1.y = vertarray[f->verts[k_cur]].y - vertarray[f->verts[k_prev]].y;
			vec1.z = vertarray[f->verts[k_cur]].z - vertarray[f->verts[k_prev]].z;

			vec2.x = vertarray[f->verts[k_cur]].x - vertarray[f->verts[k_next]].x;
			vec2.y = vertarray[f->verts[k_cur]].y - vertarray[f->verts[k_next]].y;
			vec2.z = vertarray[f->verts[k_cur]].z - vertarray[f->verts[k_next]].z;

         CrossProd(vec1, vec2, &norm);
			Normalize(&norm);

			k++;
		}

		if (k <= f->num_verts)
		{
			f->normal.x = norm.x;
			f->normal.y = norm.y;
			f->normal.z = norm.z;
			f->dist = DotProd(f->normal,vertarray[f->verts[0]]);
		}
		else
		{
			f->normal.x = 0;
			f->normal.y = 0;
			f->normal.z = 0;
			f->dist     = 0;
		}
	}
}

void RecalcAllNormals(void)
{
	brush_t *b;

	for (b=M.BrushHead; b; b=b->Next)
		RecalcNormals(b);
}


#include "edsel.h"
#include "entity.h"
#include "group.h"
#include "memory.h"
#include "s_error.h"
#include "s_gui.h"
#include "s_main.h"
#include "s_render.h"
#include "s_sys.h"
#include "uid.h"
#include "undo.h"


void Map_Unload(void)
{
   brush_t  *b,*bn;
   entity_t *e,*en;
   group_t  *g,*gn;

   Hook_Call(hk_map_unload,NULL);

   Undo_SetClear();

   ClearSelAll();

   S_R_Map_Free();

/* Must clear links before freeing or x_Free code will complain */
   for (b=M.BrushHead;b;b=bn)
   {
      bn=b->Next;
      b->Next=b->Last=NULL;
      B_Free(b);
   }
   M.BrushHead = NULL;
   M.num_brushes = 0;

   for (e=M.EntityHead;e;e=en)
   {
      en=e->Next;
      e->Next=e->Last=NULL;
      E_Free(e);
   }
   M.EntityHead = NULL;
   M.WorldSpawn = 0;
   M.num_entities = 0;

   for (g=M.GroupHead;g;g=gn)
   {
      gn=g->Next;
      g->Next=g->Last=NULL;
      G_Free(g);
   }
   M.GroupHead=NULL;
   M.WorldGroup=0;
   M.num_groups=0;
   
   M.modified=0;
   M.mapfilename[0]=0;
   S_SetMapName(NULL);

   S_T_ClearCache();

   M.valid=0;
}

void Map_New(void)
{
   Undo_SetClear();
   
   Map_Unload();

   M.mapfilename[0]=0;
   M.sel.flags=SELECTION_PRIMARY;

   Grp_CreateWorldGroup();

   M.EntityHead=E_New();
   SetKeyValue(M.EntityHead, "classname", "worldspawn");
   M.num_entities = 1;
   M.WorldSpawn=M.EntityHead->uid=GetEntityUID();
   M.EntityHead->real = 1;

   M.modified=0;
   S_SetMapName(NULL);
   M.valid=1;
   InitViewports();
   
   Hook_Call(hk_map_new,NULL);
   
   Map_Update();
}


int Map_Size(void)
{
   brush_t  *b;
   entity_t *e;
   group_t  *g;
   int size;

   size=0;

   for (b=M.BrushHead;b;b=b->Next)
      size+=sizeof(brush_t)+B_Size(b);
   for (e=M.EntityHead;e;e=e->Next)
      size+=sizeof(entity_t)+E_Size(e);
   for (g=M.GroupHead;g;g=g->Next)
      size+=sizeof(group_t)+G_Size(g);

   return size;
}

void Map_SetName(const char *name)
{
   strcpy(M.mapfilename,name);
   S_SetMapName(M.mapfilename);
}


static void map_backup(const char *name)
{ /* TODO: write plugin that uses rcs to keep backups of a map */
   char *b,*c;
   char backupname[MAX_PATH];

   strcpy(backupname,name);
   c=NULL;
   for (b=backupname;*b;b++)
   {
      if (*b=='.') c=b;
      if (*b=='/' || *b=='\\') c=NULL;
   }
   if (c) *c=0;
   strcat(backupname,".bak");
   rename(name,backupname);
}

int Map_Save(void)
{
   char name[MAX_PATH];
   
   if (M.mapfilename[0])
   {
      Hook_Call(hk_map_save,M.mapfilename);
      map_backup(M.mapfilename);
      if (!G_SaveMap(M.mapfilename))
            return 0;
      Hook_Call(hk_map_saved,M.mapfilename);
      M.modified=0;
      return 1;
   }

   if (!S_GUI_FileDialog("Save map","Save","*.map",name))
      return 0;

   if (Sys_Exists(name))
		if (!S_GUI_YesNo("Save as","File exists: '%s'\n Overwrite?",name))
         return 0;

   Map_SetName(name);
   Hook_Call(hk_map_save,M.mapfilename);
   map_backup(M.mapfilename);
   if (!G_SaveMap(M.mapfilename))
      return 0;
   Hook_Call(hk_map_saved,M.mapfilename);
   M.modified=0;
   return 1;
}


int Map_Load(const char *fname)
{
   Map_Unload();
   if (!G_LoadMap(fname))
   {
      Map_Unload();
      Map_New();
      return 0;
   }
   M.valid=1;
   M.sel.flags=SELECTION_PRIMARY;
   Map_SetName(fname);
   Grp_Update();
   Hook_Call(hk_map_load,NULL);
   Map_Update();
   M.modified=0;
   return 1;
}


void Map_UpdateVLists(void)
{ /* TODO: hooks for plugin-defined visibility changes (region, hide all that isn't selected, etc.) */
/* TODO: what if a selected brush isn't visible? */
   brush_t *b;
   entity_t *e;
   group_t *g;

   M.VBrushHead=NULL;
   for (b=M.BrushHead;b;b=b->Next)
   {
      g=UID_Group(b->Group);
      if (g && g->hidden) continue;

      b->VNext=M.VBrushHead;
      b->VLast=NULL;
      if (M.VBrushHead) M.VBrushHead->VLast=b;
      M.VBrushHead=b;
   }
   
   M.VEntityHead=NULL;
   for (e=M.EntityHead;e;e=e->Next)
   {
      g=UID_Group(e->Group);
      if (g && g->hidden) continue;

      e->VNext=M.VEntityHead;
      e->VLast=NULL;
      if (M.VEntityHead) M.VEntityHead->VLast=e;
      M.VEntityHead=e;
   }


   /* Not a great place to do this, but it's better than doing it once per
   redraw, and we can't do it in the x_Update functions or we might miss
   updating an entity if a brush's entity changed to another one. */
   {
      entity_t *e;
      int i;
      brush_t *b;

      for (e=M.VEntityHead;e;e=e->VNext)
         e->num_brushes=0;

      /* This will mess up the values for entities that aren't visible, but
      they'll be fixed once they become visible, so that doesn't matter */
      /* TODO: change that, it's better to simply guarantee that all objects
      are always up to date */
      for (b=M.VBrushHead; b; b=b->VNext)
      {
         if (!b->num_verts) continue;
         if (!(e=UID_Entity(b->Entity))) continue;
   
         e->num_brushes++;
         if (e->num_brushes==1)
         {
            e->min=b->verts[0];
            e->max=b->verts[0];
         }
      
         for (i=0;i<b->num_verts;i++)
         {
            VEC_CheckMin(e->min,b->verts[i]);
            VEC_CheckMax(e->max,b->verts[i]);
         }
      }

      for (e=M.VEntityHead;e;e=e->VNext)
      {
         if (GetKeyValue(e,"origin")!=NULL)
         {
            double cx,cy,cz;
            cx=cy=cz=0;
            sscanf(GetKeyValue(e,"origin"),"%lf %lf %lf",&cx,&cy,&cz);
            e->center.x=cx; e->center.y=cy; e->center.z=cz;
            e->min.x=cx+e->eclass->min.x;  e->min.y=cy+e->eclass->min.y;  e->min.z=cz+e->eclass->min.z;
            e->max.x=cx+e->eclass->max.x;  e->max.y=cy+e->eclass->max.y;  e->max.z=cz+e->eclass->max.z;
         }
         else
         {
            if (!e->num_brushes)
            {  /* TODO: figure out better way to handle empty entities? right now
                  they're drawn as 16x16x16 boxes at (0,0,0) */
               e->center.x=e->center.y=e->center.z=0;
               e->min.x=e->min.y=e->min.z=-8;
               e->max.x=e->max.y=e->max.z= 8;
            }
            else
            {
               e->center.x=(e->min.x+e->max.x)/2;
               e->center.y=(e->min.y+e->max.y)/2;
               e->center.z=(e->min.z+e->max.z)/2;
         
               if (e->num_brushes==1)
               {
                  e->min.x-=4;
                  e->min.y-=4;
                  e->min.z-=4;
         
                  e->max.x+=4;
                  e->max.y+=4;
                  e->max.z+=4;
               }
            }
         }
      }
   }
}

void Map_Update(void)
{
   brush_t *b;
   entity_t *e;
   group_t *g;

   for (b=M.BrushHead ;b;b=b->Next) B_Update(b);
   for (e=M.EntityHead;e;e=e->Next) E_Update(e);
   for (g=M.GroupHead ;g;g=g->Next) G_Update(g);
   Grp_Update();
   Map_UpdateVLists();
}


int Map_Close(const char *title)
{
   int i;
   
   if (!M.modified)
      return 1;
   i=S_GUI_Dialog(title,"The map has been modified since it was last saved. Save now?",
      "Save","Don't save","Cancel",NULL);
   if (i==2) return 0;
   if (i==1) return 1;

   if (!Map_Save())
      return 0;
   return 1;
}

