#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "texcat.h"

#include "memory.h"
#include "s_error.h"
#include "s_message.h"
#include "token.h"


int num_tex_category;
tex_category_t **tex_category;

void TC_Init(void)
{
   TC_Done();
}

static void TC_FreeCategory(tex_category_t *t)
{
   int i;
   if (!(t->flags&TC_DONT_FREE_NAMES))
      for (i=0;i<t->num_names;i++)
         Q_free(t->names[i]);
   Q_free(t->name);
   Q_free(t);
}

void TC_Done(void)
{
   int i;
   if (tex_category)
   {
      for (i=0;i<num_tex_category;i++)
         TC_FreeCategory(tex_category[i]);
      Q_free(tex_category);
   }
   tex_category=NULL;
   num_tex_category=0;
}


static void InsertCategory(tex_category_t *t)
{
   int i,j;
   for (i=0;i<num_tex_category;i++)
   {
      if (strcmp(t->name,tex_category[i]->name)<0)
         break;
   }
   tex_category=Q_realloc(tex_category,sizeof(tex_category_t *)*(num_tex_category+1));
   if (!tex_category)
      S_Abort("InsertCategory","Out of memory!");
   num_tex_category++;
   for (j=num_tex_category-1;j>i;j--)
      tex_category[j]=tex_category[j-1];
   tex_category[j]=t;
}


static int t_cmp(const void *v1,const void *v2)
{
   char * const *c1=v1;
   char * const *c2=v2;
   return stricmp(*c1,*c2);
}

void TC_Load(const char *filename)
{
   tex_category_t *t;
   char **names;

   if (!TokenFile(filename,TOKEN_C|TOKEN_STRING|TOKEN_ALLNAME,NULL))
   {
      S_NewMessage("Warning: Can't open texture category file '%s'!",
         filename);
      return;
   }

   while (TokenGet(1,-1))
   {
      t=Q_malloc(sizeof(tex_category_t));
      if (!t)
      {
         S_Error("TC_Load","Out of memory!");
         break;
      }
      memset(t,0,sizeof(tex_category_t));
      if (token[0]=='"')
      {
         token[strlen(token)-1]=0;
         t->name=Q_strdup(&token[1]);
      }
      else
         t->name=Q_strdup(token);
      if (!t->name)
      {
         S_Error("TC_Load","Out of memory!");
         break;
      }

      if (!TokenGet(1,-1) || strcmp(token,"{"))
      {
         S_Error("TC_Load","Parse error in '%s', line %i!",filename,token_linenum);
         break;
      }
      if (!TokenGet(1,-1))
      {
         S_Error("TC_Load","Parse error in '%s', line %i!",filename,token_linenum);
         break;
      }
      while (strcmp(token,"}"))
      {
         names=Q_realloc(t->names,sizeof(char *)*(t->num_names+1));
         if (!names)
         {
            S_Error("TC_Load","Out of memory!");
            break;
         }
         t->names=names;
         if (token[0]=='"')
         {
            token[strlen(token)-1]=0;
            t->names[t->num_names]=Q_strdup(&token[1]);
         }
         else
            t->names[t->num_names]=Q_strdup(token);
         if (!t->names[t->num_names])
         {
            S_Error("TC_Load","Out of memory!");
            return;
         }
         t->num_names++;
         if (!TokenGet(1,-1))
         {
            S_Error("TC_Load","Parse error in '%s', line %i!",filename,token_linenum);
            break;
         }
      }

      InsertCategory(t);
      qsort(t->names,t->num_names,sizeof(char *),t_cmp);
   }

   TokenDone();
}

void TC_Filter(int cat_num,int numt,const char **names,int size,unsigned char *keep)
{
/*
This assumes that both lists are sorted alphabetically without duplicates
*/
   int i,j,k;
   tex_category_t *t;
   
   if (cat_num<0 || cat_num>=num_tex_category)
   {
      S_NewMessage("Warning: TC_Filter called with invalid category number!");
      return;
   }

   t=tex_category[cat_num];
   i=j=k=0;
   do
   {
      while (!keep[i] && i<numt)
      {
         i++;
         names+=size;
      }
      if (i==numt) break;
      while (j<t->num_names && (k=stricmp(t->names[j],*names))<0)
         j++;
      if (j==t->num_names)
      {
         for (;i<numt;i++)
            keep[i]=0;
         break;
      }
      if (k!=0)
         keep[i]=0;
      else
         j++;
      names+=size;
      i++;
      if (i==numt) break;
      if (j==t->num_names)
      {
         for (;i<numt;i++)
            keep[i]=0;
         break;
      }
   } while (1);
#if 0
/* Reverse filter, useful when creating groups.
TODO: make this optional
*/
   do
   {
      while (!keep[i] && i<numt)
      {
         i++;
         names+=size;
      }
      if (i==numt) break;
      while (j<t->num_names && (k=stricmp(t->names[j],*names))<0)
         j++;
      if (j==t->num_names)
         break;
      if (k==0)
      {
         keep[i]=0;
         j++;
      }
      names+=size;
      i++;
      if (i==numt) break;
      if (j==t->num_names)
         break;
   } while (1);
#endif
}


tex_category_t *TC_New(const char *name,int flags)
{
   tex_category_t *t;

   int l,h,m;
   int i;

   if (num_tex_category)
   {
      for (l=0,h=num_tex_category-1;l<h;)
      {
         m=(l+h)/2;
         if (!(i=strcmp(name,tex_category[m]->name)))
         {
            if (flags!=tex_category[m]->flags)
               S_NewMessage("Warning: "
                  "Category '%s' with flags %i recreated with flags %i",
                  name,tex_category[m]->flags,flags);
                  
            return tex_category[m];
         }

         if (i<0) h=m-1; else l=m+1;
      }
      i=strcmp(name,tex_category[l]->name);
      if (!i)
      {
         if (flags!=tex_category[l]->flags)
            S_NewMessage("Warning: "
               "Category '%s' with flags %i recreated with flags %i",
               name,tex_category[l]->flags,flags);
                  
         return tex_category[l];
      }
      if (i>0) l++;
   }
   else
      l=0;

   t=Q_malloc(sizeof(tex_category_t));
   if (!t) S_Abort("TC_New","Out of memory!");
   memset(t,0,sizeof(tex_category_t));
   t->flags=flags;
   t->name=Q_strdup(name);
   if (!t->name) S_Abort("TC_New","Out of memory!");

   tex_category=Q_realloc(tex_category,
      sizeof(tex_category_t *)*(num_tex_category+1));
   if (!tex_category)
      S_Abort("TC_New","Out of memory!");

   for (h=num_tex_category;h>l;h--)
      tex_category[h]=tex_category[h-1];
   tex_category[h]=t;
   num_tex_category++;

   return t;
}

void TC_AddName(tex_category_t *tc,char *name)
{
   int i;
   int l,h,m;

   if (tc->num_names)
   {
      for (l=0,h=tc->num_names-1;l<h;)
      {
         m=(l+h)/2;
         if (!(i=stricmp(name,tc->names[m])))
            return;

         if (i<0) h=m-1; else l=m+1;
      }
      i=stricmp(name,tc->names[l]);
      if (!i)
         return;
      if (i>0) l++;
   }
   else
      l=0;

   tc->names=Q_realloc(tc->names,sizeof(char *)*(tc->num_names+1));
   if (!tc->names) S_Abort("TC_AddName","Out of memory!");

   for (h=tc->num_names;h>l;h--)
      tc->names[h]=tc->names[h-1];
   tc->names[h]=name;
   tc->num_names++;
}

