/*
texlock.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "texlock.h"

#include "brush.h"
#include "s_texture.h"
#include "texalign.h"


static int orientation(float p[2],float q[2],float r[2])
{
   int i;
   float a[2],b[2];

   for (i=0;i<2;i++)
   {
      a[i]=q[i]-p[i];
      b[i]=r[i]-p[i];
   }

   return (a[0]*b[0]+a[1]*b[1]>0);
}

/*
bearing of a line joining p and q points - in degrees
don't know what axis is 0, but it is not important
*/
static float angle(float p[2],float q[2])
{
   float x,y;
   float a;

   x=p[0]-q[0];
   y=p[1]-q[1];

   if (!fabs(x) && !fabs(y))
      return 0;

   if (fabs(x)>fabs(y))
   {
      a=atan(y/x);
   }
   else
   {
      a=PI/2-atan(x/y);
   }

   return a*180/PI;
}

#define Dot(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])

/*
orig plane has points and mapping parameters filled
new plane has points only filled
orig plane is the face before manipulation
new plane is the face after manipulation
only three points of the face are needed
but the points must be corresponding ones
and must not be colinear in both cases
the result of this routine is new plane's texture mapping set
*/
static void Solution(brush_t *orig_b,face_t *orig_p,brush_t *new_b,face_t *new_p)
{
/* all variants of selecting two points of three */
static int pset[3][2]={{0,1},{0,2},{1,2}};

   float orig_v[2][3];
   float  new_v[2][3],new_v1[2][3];

   int i,j;

   float p,val;

   int best;

   float orig_st[3][2];
   float orig_c[3][3];

   float new_st[3][2];
   float new_c[3][3];

   float f;

/*
calculate S and T for orig_p - you also may have these values calculated
we will not change them
*/
   if (!orig_p->normal.x &&
       !orig_p->normal.y &&
       !orig_p->normal.z)
      return;

   TextureAxisFromPlane(orig_p->normal,NULL,
      (vec3_t *)(orig_v[0]),(vec3_t *)(orig_v[1]));
   MakeVectors(orig_v,&orig_p->a);

   for (i=0;i<3;i++)
   {
      orig_c[i][0]=orig_b->verts[orig_p->verts[i]].x;
      orig_c[i][1]=orig_b->verts[orig_p->verts[i]].y;
      orig_c[i][2]=orig_b->verts[orig_p->verts[i]].z;

      orig_st[i][0]=Dot(orig_v[0],orig_c[i])+orig_p->a.shift.x;
      orig_st[i][1]=Dot(orig_v[1],orig_c[i])+orig_p->a.shift.y;
   }

// reset new_p's mapping
   new_p->a.rotate=0;
   new_p->a.scale.x=new_p->a.scale.y=1;
   new_p->a.shift.x=new_p->a.shift.y=0;

// prepare calculations for new_p
   TextureAxisFromPlane(new_p->normal,NULL,
      (vec3_t *)(new_v[0]),(vec3_t *)(new_v[1]));

   MakeVectors(new_v,&new_p->a);
   memcpy(new_v1,new_v,sizeof(new_v1));

// save them for later

// and calculate for the first time
   for (i=0;i<3;i++)
   {
      new_c[i][0]=new_b->verts[new_p->verts[i]].x;
      new_c[i][1]=new_b->verts[new_p->verts[i]].y;
      new_c[i][2]=new_b->verts[new_p->verts[i]].z;

      new_st[i][0]=Dot(new_v[0],new_c[i])+new_p->a.shift.x;
      new_st[i][1]=Dot(new_v[1],new_c[i])+new_p->a.shift.y;
   }

// first, we will resolve the rotation value
   new_p->a.rotate=angle( new_st[0], new_st[1])
                  -angle(orig_st[0],orig_st[1]);

   if (orientation(orig_st[0],orig_st[1],orig_st[2])
     ==orientation( new_st[0], new_st[1], new_st[2]))
   {
      new_p->a.rotate=180-new_p->a.rotate;
   }
/*
 it is crazy that I have to check for equal orientation
 but there are many possibilities...
 this place may be a source of some error though,
 but the idea is good
 So if wrong angles are calculated sometimes,
 check the "angle" procedure first
*/

/*
*** IMPORANT ***

If Quake doesn't handle real values for angle (e.g. whole degrees only
or 256 steps per full circle), this rounding has to be done *NOW*.
In such case, I suggest to put another parameter to this routine
which says if rounding has or has not to appear.
Then, while in program, no rounding is performed
and when saving, all mapped faces are recalculated again with rounding on.
This method minimizes texture un-alignment on manipulation
*/

/* and calculate for the second time - with known rotation */
   memcpy(new_v,new_v1,sizeof(new_v));
   MakeVectors(new_v,&new_p->a);

   for (i=0;i<3;i++)
   {
      new_st[i][0]=Dot(new_v[0],new_c[i])+new_p->a.shift.x;
      new_st[i][1]=Dot(new_v[1],new_c[i])+new_p->a.shift.y;
   }

/* now we'll resolve scale values */
   for (j=0;j<2;j++)
   {
/* for each axis select a good pair of points first */
      val=0;
      best=0;
      for (i=0;i<3;i++)
      {
         p=fabs(
              ( new_st[pset[i][0]][j]- new_st[pset[i][1]][j])
             *(orig_st[pset[i][0]][j]-orig_st[pset[i][1]][j])
               );

         if (p>val)
         {
            val=p;
            best=i;
         }
      }
/* and then calculate the scale value for given axis */

      if (val)
      {
         f=(new_st[pset[best][0]][j]- new_st[pset[best][1]][j])
          /(orig_st[pset[best][0]][j]-orig_st[pset[best][1]][j]);
      }
      else
         f=1;
      if (!j)
         new_p->a.scale.x=f;
      else
         new_p->a.scale.y=f;
   }
/*
third calculation - with known rotation and scale
only one point is necessary
*/
   memcpy(new_v,new_v1,sizeof(new_v));
   MakeVectors(new_v,&new_p->a);

   new_st[0][0]=Dot(new_v[0],new_c[0])+new_p->a.shift.x;
   new_st[0][1]=Dot(new_v[1],new_c[0])+new_p->a.shift.y;
/* find shift values */
   new_p->a.shift.x=orig_st[0][0]-new_st[0][0];
   new_p->a.shift.y=orig_st[0][1]-new_st[0][1];

/* that's all folks 8-) */

/*
after all, it is possible to calculate modulus of shift values
by texture size. Don't forget though that shift values are NOT integers.
*/
   {
      texture_t *t;

      t=S_T_LoadSize(new_p->tex.name);
      if (t)
      {
         int is;

         is=(int)(new_p->a.shift.x/(float)t->width);
         new_p->a.shift.x-=is*t->width;

         is=(int)(new_p->a.shift.y/(float)t->height);
         new_p->a.shift.y-=is*t->height;
      }
   }
}

void TexLock(brush_t *n,brush_t *o)
{
   int i;

   if (n->bt->type!=BRUSH_NORMAL) return;

   for (i=0;(i<n->num_faces) && (i<o->num_faces);i++)
   {
      Solution(o,&o->faces[i],n,&n->faces[i]);
   }
}

