/*
cmd/games/qengine/q_tex.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include "defines.h"
#include "types.h"

#include "qmap.h"

#include "game_reg.h"

#include "config.h"
#include "entclass.h"
#include "error.h"
#include "map.h"
#include "memory.h"
#include "status.h"
#include "texcat.h"
#include "token.h"

#include "fe/sys.h"
#include "fe/texture.h"
#include "fe/types.h"


/* TODO: look closely at error handling here */


static config_var_t
	*cv_default_wad,
	*cv_override_map,
	*cv_pal,
	*cv_override_pal;


typedef struct
{
	char  name[16];      /* Name of the texture */
	s_u32 width;         /* Width of texture, multiple of 8 */
	s_u32 height;        /* Height of texture, multiple of 8 */
	s_u32 offset1;       /* Offset to 1 x 1 data */
	s_u32 offset2;       /* Offset to 1/2 x 1/2 data */
	s_u32 offset4;       /* Offset to 1/4 x 1/4 data */
	s_u32 offset8;       /* Offset to 1/8 x 1/8 data */
} miptex_t;

typedef struct
{
	char magic[4];       /* "WAD2", Name of the new WAD format */
	s_u32 numentries;    /* Number of entries */
	s_u32 diroffset;     /* Position of WAD directory in file */
} wadhead_t;

typedef struct
{
	s_u32 offset;        /* Position of the entry in WAD */
	s_u32 dsize;         /* Size of the entry in WAD file */
	s_u32 size;          /* Size of the entry in memory */
	s_u8  type;          /* type of entry */
	s_u8  cmprs;         /* Compression. 0 if none. */
	s_u16 dummy;         /* Not used */
	char  name[16];      /* 1 to 16 characters, '\0'-padded */
} wadentry_t;

/*
Types:
 0x40 = Color Palette
 0x42 = Pictures for status bar
 0x44 = Mip Texture
 0x45 = Console picture (flat)
*/


typedef struct
{
	int ofs;
} g_t;

static FILE *wad;


static void LoadTex(void *tex,const texture_t *t)
{
	unsigned char *buf;
	miptex_t miptex;
	FILE *f=wad;

	if (!wad) return;

	if (fseek(f,((g_t *)t->g)->ofs,SEEK_SET))
	{
		Warning("Unable to load '%s' from the .wad: %s",t->name,strerror(errno));
		return;
	}
	if (fread(&miptex,sizeof(miptex_t),1,f)!=1)
	{
		Warning("Unable to load '%s' from the .wad: %s",t->name,strerror(errno));
		return;
	}

	S_LittleEndian(miptex.width);
	S_LittleEndian(miptex.height);
	S_LittleEndian(miptex.offset1);
	S_LittleEndian(miptex.offset2);
	S_LittleEndian(miptex.offset4);
	S_LittleEndian(miptex.offset8);

	if (fseek(f,((g_t *)t->g)->ofs+miptex.offset1,SEEK_SET))
	{
		Warning("Unable to load '%s' from the .wad: %s",t->name,strerror(errno));
		return;
	}
	buf=Q_malloc(miptex.width*miptex.height);
	if (!buf)
	{
		Warning("Unable to load '%s' from the .wad: %s",t->name,"Out of memory");
		return;
	}
	if (fread(buf,miptex.width*miptex.height,1,f)!=1)
	{
		Warning("Unable to load '%s' from the .wad: %s",t->name,strerror(errno));
		return;
	}
	S_T_AddData8(tex,miptex.width,miptex.height,0,0,buf,NULL);
	Q_free(buf);
}

static tex_funcs_t quake_tex={LoadTex};

static int LoadWad(const char *wadname,int verbose)
{
/* the currently loaded .wad file */
static char cur_wad[MAX_PATH];

	FILE *f;
	int i;
	int found_pal;

	wadhead_t head;
	wadentry_t *entry;

	unsigned char tex_pal[768];


	if (!strcmp(cur_wad,wadname))
		return 1;

	S_T_Clear();
	S_T_Init(0,&quake_tex);

	strcpy(cur_wad,wadname);

	if (wad)
	{
		fclose(wad);
		wad=NULL;
	}

	wad = f = fopen(wadname,"rb");
	if (!f)
	{
		if (verbose)
			Warning("LoadWad: Can't open '%s'!",wadname);
		return 0;
	}

#define IOERROR if (verbose) Warning("LoadWad: Error loading '%s': %s",wadname,strerror(errno)); fclose(f); wad=NULL;

	if (fread(&head,sizeof(wadhead_t),1,f)!=1)
	{
		IOERROR
		return 0;
	}

	if ((head.magic[0]!='W') ||
		 (head.magic[1]!='A') ||
		 (head.magic[2]!='D') ||
		 (head.magic[3]!='2'))
	{
		if (verbose)
			Warning("LoadWad: '%s' doesn't have a valid WAD2 header!",wadname);
		fclose(f);
		wad=NULL;
		return 0;
	}
	S_LittleEndian(head.diroffset);
	S_LittleEndian(head.numentries);

	if (fseek(f,head.diroffset,SEEK_SET))
	{
		IOERROR
		return 0;
	}

	entry = (wadentry_t *)Q_malloc(sizeof(wadentry_t) * head.numentries);
	if (!entry)
	{
		if (verbose)
			Warning("LoadWad: Out of memory!");
		fclose(f);
		wad=NULL;
		return 0;
	}
	if (fread(entry, sizeof(wadentry_t) * head.numentries, 1, f)!=1)
	{
		IOERROR
		Q_free(entry);
		return 0;
	}

	found_pal=0;
	for (i=0;i<head.numentries;i++)
	{
		S_LittleEndian(entry[i].offset);
		S_LittleEndian(entry[i].dsise);
		S_LittleEndian(entry[i].size);

		if (!cv_override_pal->i)
		{
			switch (entry[i].type)
			{
			case 0x40:
				if (fseek(f,entry[i].offset,SEEK_SET))
				{
					IOERROR
					Q_free(entry);
					return 0;
				}
				if (fread(tex_pal, sizeof(tex_pal), 1, f)!=1)
				{
					IOERROR
					Q_free(entry);
					return 0;
				}
				S_T_SetPalette(tex_pal);
				found_pal=1;
				break;
			}
		}
	}

	if (!found_pal)
	{
		FILE *f2;

		f2=Sys_OpenFile(cv_pal->s,"rb");
		if (!f2)
		{
			Warning("LoadWad: No palette in .wad file, and couldn't open '%s'. Reverting to grey palette.",
				cv_pal->s);
			for (i=0;i<256;i++)
				 tex_pal[i*3]=tex_pal[i*3+1]=tex_pal[i*3+2]=(i&15)*17;
		}
		else
		{
			fread(tex_pal,sizeof(tex_pal),1,f2);
			fclose(f2);
		}
		S_T_SetPalette(tex_pal);
	}

	for (i=0;i<head.numentries;i++)
	{
		if (entry[i].type!=0x44) continue;

		{
			g_t *g;
			struct {s_u32 x; s_u32 y; } h;

			g=Q_malloc(sizeof(g_t));
			if (!g) return 0;
			g->ofs=entry[i].offset;
			strlwr(entry[i].name);

			if (fseek(f,entry[i].offset+16,SEEK_SET))
			{
				IOERROR
				Q_free(entry);
				return 0;
			}
			if (fread(&h,sizeof(h),1,f)!=1)
			{
				IOERROR
				Q_free(entry);
				return 0;
			}

			S_LittleEndian(h.x);
			S_LittleEndian(h.y);

			if (!S_T_AddTexture(entry[i].name,h.x,h.y,g))
				return 0;
		}
	}

	Q_free(entry);

	return 1;
#undef IOERROR
}

/* TODO
static int Q_TexBSPFlags(texdef_t *t)
{
   if (t->name[0]=='*')             return TEX_FULLBRIGHT|TEX_NONSOLID;
   if (!strnicmp(t->name,"SKY",3))  return TEX_FULLBRIGHT;
   if (!stricmp(t->name,"TRIGGER")) return TEX_NODRAW;
   if (!stricmp(t->name,"CLIP"))    return TEX_NODRAW|TEX_NONSOLID;
   return 0;
}
*/

typedef struct G_texdef_s
{
	int detail;
} G_texdef_t;


static int QT_Init(texdef_t *t)
{
	if (!t->g)
	{
		t->g=Q_malloc(sizeof(G_texdef_t));
		if (!t->g) return 0;
	}
	t->g->detail=0;
	return 1;
}

static void QT_Free(texdef_t *t)
{
	if (t->g)
	{
		Q_free(t->g);
		t->g=NULL;
	}
}

static int QT_Copy(texdef_t *dst,const texdef_t *src)
{
	if (!src->g) return 1;
	if (!dst->g)
	{
		dst->g=Q_malloc(sizeof(G_texdef_t));
		if (!dst->g) return 0;
	}
	dst->g->detail=src->g->detail;
	return 1;
}


static game_reg_t G_Quake=
{
	"Quake",
	{
		QMap_Load,
		QMap_Save
	},
	{
		QT_Init,
		QT_Free,
		QT_Copy
	},
};


static int Q_LoadTexInfo(texdef_t *t)
{
	if (!t->g)
	{
		t->g=Q_malloc(sizeof(G_texdef_t));
		if (!t->g) return ERROR_NOMEM;
	}
	t->g->detail=0;

	while (TokenAvailable(0))
	{
		GETTOKEN(0,TOKEN_NAME);
		if (!strcmp(token,"detail"))
			t->g->detail=1;
		else
//			; /* TODO */
			return ERROR_PARSE;
	}

	return ERROR_NO;
}

static void Q_SaveTexInfo(texdef_t *t,FILE *fp)
{
	if (t->g && t->g->detail)
		fprintf(fp," detail");
}


#include "qfunc.h"

#include "entity.h"
#include "hook.h"
#include "quest.h"
#include "uid.h"

#include "fe/gui.h"
#include "fe/message.h"

/* TODO: set default? */
static void C_game_sys_quake1_select_wad(void)
{
	char filename[MAX_PATH];

	filename[0]=0;
	if (!S_GUI_FileDialog("Select .wad","Load","*.wad",filename))
		return;

	SetKeyValue(UID_Entity(M.WorldSpawn),"wad",filename);
	LoadWad(filename,1);
	Map_Update();
	S_NewMessage("Wadfile set to: %s",filename);
}
static func_t F_game_sys_quake1_select_wad=
{"game.sys.quake1.select_wad",FUNC_VOID,{fv:C_game_sys_quake1_select_wad},NULL,NULL,0,
{"Select a .wad file to use for textures."}
};

static func_t *funcs[]={&F_game_sys_quake1_select_wad};


static void q1_map_new(void)
{
	entity_t *e=UID_Entity(M.WorldSpawn);
	if (e)
		SetKeyValue(e,"wad",cv_default_wad->s);
	LoadWad(cv_default_wad->s,1);
}

static void q1_map_load(void)
{
	entity_t *e=UID_Entity(M.WorldSpawn);
	const char *c;
	if (!e) return;
	c=GetKeyValue(e,"wad");
	if (cv_override_map->i || !c || !Sys_Exists(c) || !LoadWad(c,1))
		LoadWad(cv_default_wad->s,1);
}


#include "plugin.h"
PLUGIN_HEAD

static int Done(void);

static plugin_t plugin=
{{Init,Done},{"game.sys.quake1","Game plugin for Quake I and games/mods based on it (e.g. Hexen 2)."}};

static int Init(void)
{
	cv_default_wad=Cfg_DefVar(plugin.cfg,
		"game.sys.quake1.default_wad",
		NULL,
		"Path to the default .wad file to load textures from.");

	cv_override_map=Cfg_DefVar(plugin.cfg,
		"game.sys.quake1.override_map",
		"0",
		"If 1, Quest will always load textures from the default .wad file "
		"even if a map specifies its own .wad file.");

	cv_pal=Cfg_DefVar(plugin.cfg,
		"game.sys.quake1.pal",
		NULL,
		"Path to the default palette to use if a .wad file doesn't contain "
		"one.");

	cv_override_pal=Cfg_DefVar(plugin.cfg,
		"game.sys.quake1.override_pal",
		"0",
		"If 1, Quest will always use the palette specified in "
		"'game.sys.quake1.pal', even if a .wad file contains its own palette. "
		"If 0, Quest will load the palette from the .wad file if it contains "
		"one, and revert to 'game.sys.quake1.pal' if it doesn't.");


/* TODO: check for other game before unloading map */
	if (!Map_Close("Change game"))
		return 0;
	Map_Unload();
		
	if (!G_Load(&G_Quake))
	{
		Warning("game.quake1: Can't load, another game is already loaded!");
		return 0;
	}

	QMap_Init(plugin.cfg);
	
	qmap_loadtexinfo=Q_LoadTexInfo;
	qmap_savetexinfo=Q_SaveTexInfo;

	Func_Register(&plugin,funcs,sizeof(funcs));
/* TODO
	Hook_AddFunc(hk_map_new,HOOK_FUNC_VOID,(hook_func_void_t)q1_map_new,NULL,&plugin);
	Hook_AddFunc(hk_map_load,HOOK_FUNC_VOID,(hook_func_void_t)q1_map_load,NULL,&plugin);
*/

	EClass_Reset(); /* TODO */
	TC_Reset();

	Map_New();
	q1_map_new(); /* TODO */
	return 1;
}

static int Done(void)
{
	if (!Map_Close("Unload Quake 1"))
		return 0;
	Map_Unload();
	G_Unload();
	S_T_Clear();
	EClass_Reset();
	TC_Reset();
	Map_New();
	return 1;
}

PLUGIN_FOOT

