#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <GL/gl.h>

#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>

#include "quest/defines.h"
#include "quest/types.h"

#include "quest/fe/texpick.h"
#include "g_texpick.h"

#include "quest/fe/texture.h"
#include "../../gl/gl_texture.h"

#include "g_3d_all.h"

#include "../gui.h"

#include "quest/config.h"
#include "quest/error.h"
#include "quest/memory.h"
#include "quest/ql.h"
#include "quest/texcat.h"


static config_var_t
   *cv_gtk_gl_texpick_layout,
   *cv_gtk_gl_texpick_scale,
   *cv_gtk_gl_texpick_min_size,
   *cv_gtk_gl_texpick_max_size;


typedef struct
{
   const char *name;
   gl_texture_t *t;
   int x1,y1,x2,y2;
   int used;
} ptex_t;


static ptex_t *pt;
static int num_pt;

static ptex_t *sel;

static unsigned char *sel_groups;
static int sel_num_groups;

static int cache_only,add_mode;


static void FreeTexList(void)
{
   Q_free(pt);
   pt=NULL;
   num_pt=0;
   sel=NULL;
}

static void GetTexList(void)
{
   gl_texture_t **tex;
   int num_tex;
   int i;

   if (pt) FreeTexList();

   S_GL_GetTexs(&tex,&num_tex);

   pt=Q_malloc(sizeof(ptex_t)*num_tex);
   if (!pt)
   {
      Warning("Texture picker: Out of memory!");
      pt=NULL;
      num_pt=0;
      return;
   }
   memset(pt,0,sizeof(ptex_t)*num_tex);

   num_pt=0;
   for (i=0;i<num_tex;i++)
   {
      if (cache_only && !tex[i]->tex.cached)
         continue;
      pt[num_pt].t=tex[i];
      pt[num_pt].name=pt[num_pt].t->tex.name;
      num_pt++;
   }

   if (num_pt)
   {
      unsigned char *keep;
      int delta;
      
      keep=Q_malloc(num_pt);
      if (!keep)
      {
         Warning("Texture picker: Out of memory!");
         return;
      }
      memset(keep,1,num_pt);

      for (i=0;i<sel_num_groups;i++)
         if (sel_groups[i] && i!=0)
            TC_Filter(i-1,num_pt,&pt[0].name,sizeof(ptex_t)/sizeof(char *),keep);
   
      for (delta=i=0;i<num_pt;i++)
      {
         if (!keep[i])
            delta++;
         else
         if (delta)
            pt[i-delta]=pt[i];
      }
      num_pt-=delta;
      Q_free(keep);
   }
   
   for (i=0;i<num_pt;i++)
      S_GL_WantLoad(pt[i].t,1);
}


static ptex_t *FindPT(int x,int y)
{
   ptex_t *p;
   int i;

   for (i=num_pt,p=pt;i;i--,p++)
   {
      if (x<p->x1 || x>p->x2 || y<p->y1 || y>p->y2)
         continue;
      return p;
   }
   return NULL;
}


static ptex_t *FindName(const char *name)
{
   ptex_t *p;
   int i;
   for (i=num_pt,p=pt;i;i--,p++)
      if (!stricmp(name,p->t->tex.name))
         return p;
   return NULL;
}


static int scroll_x,scroll_y;
static int smax_x,smax_y;
static int t_x,t_y;

static GtkGLArea *t_gla;


/* Different flow methods */
static void flow_xa(void)
{
   int curx,cury;
   int maxx;
   ptex_t *p;
   int i;
   
   curx=cury=4;
   maxx=0;
   smax_y=0;

   for (i=0,p=pt;i<num_pt;i++,p++)
   {
      if (p->y2+cury+4>t_y)
      {
         curx+=maxx+4;
         cury=4;
         maxx=0;
      }

      if (p->x2>maxx)
         maxx=p->x2;

      p->x1+=curx;
      p->y1+=cury;
      p->x2+=curx;
      p->y2+=cury;

      cury=p->y2+4;
      
      if (cury>smax_y) smax_y=cury;
   }
   smax_x=curx+maxx+4;
}

static void flow_xs(void)
{
   ptex_t *p,*p2,*p3;
   int i,j,k;
   int goalx;
   int curx,cury,maxx;

   curx=cury=4;
   maxx=0;
   smax_y=0;

   for (i=0,p=pt;i<num_pt;i++,p++)
   {
      if (p->used) continue;

      goalx=0;

      for (p2=p,j=i;j<num_pt;j++,p2++)
      {
         if (p2->used) continue;

         if (!goalx)
            goalx=p2->x2;
         else
         {
            if (p2->x2<goalx*3/4 || p2->x2>goalx*5/4)
               continue;
            if (p2->y2+cury+4>t_y)
               continue;
         }

         if (p2->x2>maxx)
            maxx=p2->x2;
            
         p2->x1+=curx;
         p2->y1+=cury;
         p2->x2+=curx;
         p2->y2+=cury;

         cury=p2->y2+4;
      
         if (cury>smax_y) smax_y=cury;
         p2->used=1;
      }

		goalx=maxx/2;
		/* search for pairs of ~maxx/2 with similar heights */
		/* TODO: add similar code to flow_ys() ? */
		for (p2=p,j=i;j<num_pt;j++,p2++)
		{
			if (p2->used) continue;

			if (p2->x2<goalx*3/4 || p2->x2>goalx)
				continue;

			if (p2->y2+cury+4>t_y)
				continue;

			for (p3=p2+1,k=j+1;k<num_pt;k++,p3++)
			{
				if (p3->used)
					continue;

				if (p3->x2<goalx*3/4 || p3->x2>goalx)
					continue;

				if (p3->y2+cury+4>t_y)
					continue;

				break;
			}
			if (k==num_pt) continue;

			p3->x1+=curx+p2->x2+1;
			p3->y1+=cury;
			p3->x2+=curx+p2->x2+1;
			p3->y2+=cury;

			p2->x1+=curx-1;
			p2->y1+=cury;
			p2->x2+=curx-1;
			p2->y2+=cury;

			if (p2->y2>p3->y2)
				cury=p2->y2+4;
			else
				cury=p3->y2+4;

			if (cury>smax_y) smax_y=cury;
			p2->used=1;
			p3->used=1;
		}


      curx+=maxx+4;
      cury=4;
      maxx=0;
   }
   smax_x=curx+maxx+4;
}


static void flow_ya(void)
{
   int curx,cury;
   int maxy;
   ptex_t *p;
   int i;
   
   curx=cury=4;
   maxy=0;
   smax_x=0;

   for (i=0,p=pt;i<num_pt;i++,p++)
   {
      if (p->x2+curx+4>t_x)
      {
         cury+=maxy+4;
         curx=4;
         maxy=0;
      }

      if (p->y2>maxy)
         maxy=p->y2;

      p->x1+=curx;
      p->y1+=cury;
      p->x2+=curx;
      p->y2+=cury;

      curx=p->x2+4;
      
      if (curx>smax_x) smax_x=curx;
   }
   smax_y=cury+maxy+4;
}

static void flow_ys(void)
{
   ptex_t *p,*p2;
   int i,j;
   int goaly;
   int curx,cury,maxy;

   curx=cury=4;
   maxy=0;
   smax_x=0;

   for (i=0,p=pt;i<num_pt;i++,p++)
   {
      if (p->used) continue;

      goaly=0;

      for (p2=p,j=i;j<num_pt;j++,p2++)
      {
         if (p2->used) continue;

         if (!goaly)
            goaly=p2->y2;
         else
         {
            if (p2->y2<goaly*0.75 || p2->y2>goaly*1.25)
               continue;
            if (p2->x2+curx+4>t_x)
               break;
         }

         if (p2->y2>maxy)
            maxy=p2->y2;
            
         p2->x1+=curx;
         p2->y1+=cury;
         p2->x2+=curx;
         p2->y2+=cury;

         curx=p2->x2+4;
      
         if (curx>smax_x) smax_x=curx;
         p2->used=1;
      }
      cury+=maxy+4;
      curx=4;
      maxy=0;
   }
   smax_y=cury+maxy+4;
}


static void reflow(void)
{
   int i;
   ptex_t *p;

   double scale;

   for (i=0,p=pt;i<num_pt;i++,p++)
   {
      p->used=0;

      scale=cv_gtk_gl_texpick_scale->f;
      if (cv_gtk_gl_texpick_min_size->i)
      {
         if (scale*p->t->tex.width<cv_gtk_gl_texpick_min_size->i)
            scale=cv_gtk_gl_texpick_min_size->i/(double)p->t->tex.width;
         if (scale*p->t->tex.height<cv_gtk_gl_texpick_min_size->i)
            scale=cv_gtk_gl_texpick_min_size->i/(double)p->t->tex.height;
      }
      if (cv_gtk_gl_texpick_max_size->i)
      {
         if (scale*p->t->tex.width>cv_gtk_gl_texpick_max_size->i)
            scale=cv_gtk_gl_texpick_max_size->i/(double)p->t->tex.width;
         if (scale*p->t->tex.height>cv_gtk_gl_texpick_max_size->i)
            scale=cv_gtk_gl_texpick_max_size->i/(double)p->t->tex.height;
      }

      p->x2=p->t->tex.width*scale;
      p->y2=p->t->tex.height*scale;
      p->x1=p->y1=0;
   }

   switch (cv_gtk_gl_texpick_layout->i)
   {
   default:
   case 0: flow_xa(); break;
   case 1: flow_xs(); break;
   case 2: flow_ya(); break;
   case 3: flow_ys(); break;
   }
}


static void fix_position(void)
{
   if (scroll_x+t_x>smax_x) scroll_x=smax_x-t_x;
   if (scroll_x<0) scroll_x=0;
   if (scroll_y+t_y>smax_y) scroll_y=smax_y-t_y;
   if (scroll_y<0) scroll_y=0;
}


static void draw(void)
{
   int i;
   ptex_t *p;
   int sx,sy;
   int ex,ey;

   if (!gtk_gl_area_make_current(t_gla))
      return;

   glClear(GL_COLOR_BUFFER_BIT);
   glColor4f(1,1,1,1);

   sx=scroll_x;
   sy=scroll_y;
   ex=sx+t_x;
   ey=sy+t_y;
   glEnable(GL_TEXTURE_2D);
   for (i=num_pt,p=pt;i;i--,p++)
   {
      if (p->x1>ex || p->y1>ey ||
          p->x2<sx || p->y2<sy)
         continue;

      if (p->t->load<2)
         if (!S_GL_WantLoad(p->t,2))
            continue;

      glBindTexture(GL_TEXTURE_2D,p->t->num);
      glBegin(GL_QUADS);
      glTexCoord2f(0,0);
      glVertex2f(p->x1-sx,p->y1-sy);
      glTexCoord2f(1,0);
      glVertex2f(p->x2-sx,p->y1-sy);
      glTexCoord2f(1,1);
      glVertex2f(p->x2-sx,p->y2-sy);
      glTexCoord2f(0,1);
      glVertex2f(p->x1-sx,p->y2-sy);
      glEnd();
   }
   glBindTexture(GL_TEXTURE_2D,0);
   glDisable(GL_TEXTURE_2D);
   
   if (S_tex_use_cache && !cache_only)
   {
      glColor4f(1,1,0,1);
      for (i=num_pt,p=pt;i;i--,p++)
      {
         if (!p->t->tex.cached)
            continue;
      
         if (p->x1>ex || p->y1>ey ||
             p->x2<sx || p->y2<sy)
            continue;

         glBegin(GL_LINE_LOOP);
         glVertex2f(p->x1-sx-1,p->y1-sy-1);
         glVertex2f(p->x2-sx+1,p->y1-sy-1);
         glVertex2f(p->x2-sx+1,p->y2-sy+1);
         glVertex2f(p->x1-sx-1,p->y2-sy+1);
         glEnd();
      }
   }
   
   if (sel)
   {
      glColor4f(1,1,1,1);
      glBegin(GL_LINE_LOOP);
      glVertex2f(sel->x1-sx,sel->y1-sy);
      glVertex2f(sel->x2-sx,sel->y1-sy);
      glVertex2f(sel->x2-sx,sel->y2-sy);
      glVertex2f(sel->x1-sx,sel->y2-sy);
      glEnd();
   }
   
   gtk_gl_area_swap_buffers(t_gla);
}

static gint t_configure(GtkWidget *w,ARG_UNUSED(GdkEventConfigure *e),ARG_UNUSED(gpointer data))
{
   if (t_x==w->allocation.width && t_y==w->allocation.height)
      return TRUE;

   t_x=w->allocation.width;
   t_y=w->allocation.height;
   reflow();

   fix_position();

   if (!gtk_gl_area_make_current(t_gla))
      return TRUE;
   
   glViewport(0,0,t_x,t_y);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0,t_x,t_y,0,-1,1);
   
   glClearColor(0.3,0.3,0.3,0);
   draw();
   return TRUE;
}

static gint t_expose(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEventExpose *e),ARG_UNUSED(gpointer data))
{
   draw();
   return FALSE;
}


static GtkEntry *tn;
static int status;
static GtkLabel *label_name,*label_desc;


static int t_motion(ARG_UNUSED(GtkWidget *w),GdkEventMotion *e)
{
   GdkModifierType s;
   int x,y;
static int oldx,oldy;
   int dx,dy;

   gdk_window_get_pointer(e->window,&x,&y,&s);
         
   dx=x-oldx;
   dy=y-oldy;
   oldx=x;
   oldy=y;

   if (s&GDK_BUTTON3_MASK)
   {
      scroll_x-=dx;
      if (scroll_x+t_x>smax_x) scroll_x=smax_x-t_x;
      if (scroll_x<0) scroll_x=0;
      scroll_y-=dy;
      if (scroll_y+t_y>smax_y) scroll_y=smax_y-t_y;
      if (scroll_y<0) scroll_y=0;
      draw();
   }

   {
      ptex_t *p;

      p=FindPT(x+scroll_x,y+scroll_y);
      if (p)
      {
         char buf[256];
         snprintf(buf,sizeof(buf),"%3ix%3i  %s",
            p->t->tex.width,p->t->tex.height,
            p->t->tex.name);
         gtk_label_set_text(label_name,buf);
         S_T_GetDescription(buf,&p->t->tex);
         gtk_label_set_text(label_desc,buf);
      }
      else
      {
         gtk_label_set_text(label_name,NULL);
         gtk_label_set_text(label_desc,NULL);
      }
   }
   
   return 0;
}

static gint t_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),ARG_UNUSED(gpointer data))
{
   status=1;
   gtk_main_quit();
   return TRUE;
}

static gint t_mouse(ARG_UNUSED(GtkWidget *w),GdkEventButton *e,ARG_UNUSED(gpointer data))
{
   ptex_t *p;
   int x,y;
   GdkModifierType s;

   gdk_window_get_pointer(e->window,&x,&y,&s);
   p=FindPT(x+scroll_x,y+scroll_y);

   if (!p) return TRUE;

   switch (e->button)
   {
   case 1:
      if (add_mode)
      {
         if (p->t->tex.cached)
         {
            S_T_Uncache(p->t->tex.name);
            if (cache_only)
            {
               GetTexList();
               reflow();
            }
         }
         else
            S_T_Cache(p->t->tex.name);
      }
      else
      {
         sel=p;
         gtk_entry_set_text(tn,p->t->tex.name);
         if (e->type==GDK_2BUTTON_PRESS)
         {
            status=0;
            gtk_main_quit();
         }
      }
      draw();
      break;
   case 2:
      if (S_tex_use_cache)
      {
         if (p->t->tex.cached)
         {
            S_T_Uncache(p->t->tex.name);
            if (cache_only)
            {
               GetTexList();
               reflow();
            }
         }
         else
            S_T_Cache(p->t->tex.name);
      }
      draw();
      break;
   }

   return TRUE;
}

static void tn_act(ARG_UNUSED(GtkEditable *e),GtkContainer *c)
{
   gtk_container_focus(c,GTK_DIR_TAB_FORWARD);
}

static void tn_ch(ARG_UNUSED(GtkEntry *e),ARG_UNUSED(gpointer data))
{
}


static void t_click_ok(ARG_UNUSED(GtkButton *b),ARG_UNUSED(gpointer data))
{
   status=0;
   gtk_main_quit();
}

static void t_click_cancel(ARG_UNUSED(GtkButton *b),ARG_UNUSED(gpointer data))
{
   status=1;
   gtk_main_quit();
}


static void t_cache_toggle(GtkToggleButton *b)
{
   int t;

   t=gtk_toggle_button_get_active(b);
   if (t!=cache_only)
   {
      cache_only=t;
      GetTexList();
      reflow();
      fix_position();
      draw();
   }
}

static void t_add_toggle(GtkToggleButton *b)
{
   add_mode=gtk_toggle_button_get_active(b);
}


static GtkWindow *gwin;
static GtkList *glist;

static void tg_change(GtkList *list)
{
   GList *g;
   GtkWidget *w;

   memset(sel_groups,0,sel_num_groups);
   for (g=list->selection;g;g=g->next)
   {
      w=GTK_WIDGET(g->data);
      sel_groups[gtk_list_child_position(list,w)]=1;
   }
   GetTexList();
   reflow();
   fix_position();
   draw();
}

static gint tg_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),ARG_UNUSED(gpointer data))
{
   S_SaveWinPosition(gwin,"texpick_groups");
   gtk_signal_disconnect_by_func(GTK_OBJECT(glist),GTK_SIGNAL_FUNC(tg_change),NULL);
   return FALSE;
}

static void t_click_groups(ARG_UNUSED(GtkButton *b),ARG_UNUSED(gpointer data))
{
   GtkWidget *li;
   GtkScrolledWindow *sw;
   int i;

   gwin=GTK_WINDOW(gtk_window_new(GTK_WINDOW_DIALOG));
   gtk_window_set_title(gwin,"Groups");
   gtk_container_set_border_width(GTK_CONTAINER(gwin),4);
   gtk_signal_connect(GTK_OBJECT(gwin),"delete_event",GTK_SIGNAL_FUNC(tg_delete),NULL);
   gtk_window_set_modal(gwin,TRUE);

   sw=GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(NULL,NULL));
   gtk_container_add(GTK_CONTAINER(gwin),GTK_WIDGET(sw));

   glist=GTK_LIST(gtk_list_new());
   gtk_list_set_selection_mode(glist,GTK_SELECTION_MULTIPLE);
   gtk_scrolled_window_add_with_viewport(sw,GTK_WIDGET(glist));

   li=gtk_list_item_new_with_label("[All]");
   gtk_container_add(GTK_CONTAINER(glist),li);
   gtk_widget_show(li);

   for (i=0;i<num_tex_category;i++)
   {
      li=gtk_list_item_new_with_label(tex_category[i]->name);
      gtk_container_add(GTK_CONTAINER(glist),li);
      gtk_widget_show(li);
   }

   for (i=0;i<sel_num_groups;i++)
      if (sel_groups[i])
         gtk_list_select_item(glist,i);
      
   gtk_signal_connect(GTK_OBJECT(glist),"selection-changed",GTK_SIGNAL_FUNC(tg_change),NULL);

   gtk_widget_show(GTK_WIDGET(glist));
   gtk_widget_show(GTK_WIDGET(sw));
   gtk_widget_show(GTK_WIDGET(gwin));
   S_LoadWinPosition(gwin,"texpick_groups");
}


int S_TexturePicker(char *dest,const char *begin)
{
   GtkWindow *win;
   GtkTable *tb;
   GtkGLArea *gla;
   GtkWidget *w;
   GtkBox *bb;

   win=GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
   gtk_window_set_title(win,"Texture picker");
   gtk_window_set_modal(win,TRUE);
   gtk_container_set_border_width(GTK_CONTAINER(win),4);
   gtk_signal_connect(GTK_OBJECT(win),"delete_event",GTK_SIGNAL_FUNC(t_delete),NULL);

   tb=GTK_TABLE(gtk_table_new(7,1,FALSE));
   gtk_container_add(GTK_CONTAINER(win),GTK_WIDGET(tb));

   {
static int attribs[]=
{
GDK_GL_RGBA,
GDK_GL_DOUBLEBUFFER,
GDK_GL_RED_SIZE,1,
GDK_GL_GREEN_SIZE,1,
GDK_GL_BLUE_SIZE,1,
GDK_GL_NONE
};
      t_gla=gla=GTK_GL_AREA(gtk_gl_area_share_new(attribs,S_cur_gvport->gla));
      if (!gla)
      {
         Error_Add("TexturePicker: Couldn't create OpenGL context!");
         return 0;
      }
   }

   gtk_window_set_default_size(win,300,300);
   gtk_window_set_position(win,GTK_WIN_POS_MOUSE);
   
   gtk_widget_set_events(GTK_WIDGET(gla),
      GDK_BUTTON_PRESS_MASK|
      GDK_EXPOSURE_MASK|
      GDK_POINTER_MOTION_MASK|GDK_POINTER_MOTION_HINT_MASK);
      
   label_name=GTK_LABEL(gtk_label_new(NULL));
   gtk_misc_set_alignment(GTK_MISC(label_name),0.0,0.5);
   
   label_desc=GTK_LABEL(gtk_label_new(NULL));
   gtk_misc_set_alignment(GTK_MISC(label_desc),0.0,0.5);
   
   gtk_signal_connect(GTK_OBJECT(gla),"button_press_event",GTK_SIGNAL_FUNC(t_mouse),NULL);
   gtk_signal_connect(GTK_OBJECT(gla),"motion_notify_event",GTK_SIGNAL_FUNC(t_motion),NULL);

   gtk_signal_connect(GTK_OBJECT(gla),"configure_event",GTK_SIGNAL_FUNC(t_configure),NULL);
   gtk_signal_connect(GTK_OBJECT(gla),"expose_event",GTK_SIGNAL_FUNC(t_expose),NULL);

   gtk_table_attach(tb,GTK_WIDGET(gla),0,1,0,1,
      GTK_FILL|GTK_EXPAND|GTK_SHRINK,
      GTK_FILL|GTK_EXPAND|GTK_SHRINK,
      2,2);


   gtk_table_attach(tb,GTK_WIDGET(label_name),0,1,1,2,GTK_FILL|GTK_SHRINK|GTK_EXPAND,0,2,2);
   gtk_widget_show(GTK_WIDGET(label_name));
   
   gtk_table_attach(tb,GTK_WIDGET(label_desc),0,1,2,3,GTK_FILL|GTK_SHRINK|GTK_EXPAND,0,2,2);
   gtk_widget_show(GTK_WIDGET(label_desc));

   w=gtk_hseparator_new();
   gtk_table_attach(tb,w,0,1,3,4,GTK_FILL|GTK_SHRINK|GTK_EXPAND,0,2,2);
   gtk_widget_show(w);

   w=gtk_label_new("Texture name:");
   gtk_misc_set_alignment(GTK_MISC(w),0.0,0.5);
   gtk_table_attach(tb,w,0,1,4,5,GTK_FILL|GTK_SHRINK|GTK_EXPAND,0,2,2);
   gtk_widget_show(w);
   
   tn=GTK_ENTRY(gtk_entry_new());
   gtk_entry_set_text(tn,begin?begin:"");
   gtk_signal_connect(GTK_OBJECT(tn),"activate",GTK_SIGNAL_FUNC(tn_act),tb);
   gtk_signal_connect(GTK_OBJECT(tn),"changed",GTK_SIGNAL_FUNC(tn_ch),NULL);
   gtk_table_attach(tb,GTK_WIDGET(tn),0,1,5,6,GTK_FILL|GTK_SHRINK|GTK_EXPAND,0,2,2);
   gtk_widget_show(GTK_WIDGET(tn));

   w=gtk_hseparator_new();
   gtk_table_attach(tb,w,0,1,6,7,GTK_FILL|GTK_SHRINK|GTK_EXPAND,0,2,2);
   gtk_widget_show(w);

   bb=GTK_BOX(gtk_hbox_new(FALSE,4));
   gtk_table_attach(tb,GTK_WIDGET(bb),0,1,7,8,GTK_FILL|GTK_SHRINK|GTK_EXPAND,0,2,2);

   w=gtk_button_new_with_label("OK");
   gtk_signal_connect(GTK_OBJECT(w),"clicked",GTK_SIGNAL_FUNC(t_click_ok),NULL);
   gtk_box_pack_start(bb,w,0,0,2);
   gtk_widget_show(w);
   
   w=gtk_button_new_with_label("Cancel");
   gtk_signal_connect(GTK_OBJECT(w),"clicked",GTK_SIGNAL_FUNC(t_click_cancel),NULL);
   gtk_box_pack_start(bb,w,0,0,2);
   gtk_widget_show(w);

   w=gtk_button_new_with_label("Groups");
   gtk_signal_connect(GTK_OBJECT(w),"clicked",GTK_SIGNAL_FUNC(t_click_groups),NULL);
   gtk_box_pack_start(bb,w,0,0,2);
   gtk_widget_show(w);

   if (S_tex_use_cache)
   {
      w=gtk_check_button_new_with_label("Cache only");
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w),1);
      gtk_signal_connect(GTK_OBJECT(w),"toggled",GTK_SIGNAL_FUNC(t_cache_toggle),NULL);
      gtk_box_pack_start(bb,w,0,0,2);
      gtk_widget_show(w);
      
      w=gtk_check_button_new_with_label("Add mode");
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w),0);
      gtk_signal_connect(GTK_OBJECT(w),"toggled",GTK_SIGNAL_FUNC(t_add_toggle),NULL);
      gtk_box_pack_start(bb,w,0,0,2);
      gtk_widget_show(w);
      
      cache_only=1;
   }
   else
      cache_only=0;
   
   add_mode=0;

   if (sel_num_groups!=num_tex_category+1)
   {
      sel_num_groups=num_tex_category+1;
      sel_groups=Q_realloc(sel_groups,sel_num_groups);
      if (!sel_groups)
         ABORT("Texture picker: Out of memory!");
      memset(sel_groups,0,sel_num_groups);
   }
   
   GetTexList();
   scroll_x=scroll_y=0;
   t_x=t_y=-1;
   reflow();

   gtk_widget_show(GTK_WIDGET(bb));
   
   gtk_widget_show(GTK_WIDGET(gla));
   gtk_widget_show(GTK_WIDGET(tb));
   gtk_widget_show(GTK_WIDGET(win));
   S_LoadWinPosition(win,"texpick");

   if (begin)
   {
      sel=FindName(begin);
      if (sel)
      {
         scroll_x=sel->x1-t_x/2;
         scroll_y=sel->y1-t_y/2;
         if (scroll_x+t_x>smax_x) scroll_x=smax_x-t_x;
         if (scroll_x<0) scroll_x=0;
         if (scroll_y+t_y>smax_y) scroll_y=smax_y-t_y;
         if (scroll_y<0) scroll_y=0;
      }
   }
   else
      sel=NULL;

   draw();
         
   status=0;
   gtk_main();

   if (!status)
   {
      strcpy(dest,gtk_entry_get_text(tn));
      S_T_Cache(dest);
   }

   S_SaveWinPosition(win,"texpick");
   gtk_widget_destroy(GTK_WIDGET(win));

   FreeTexList();
   
   return !status;
}


void S_InitTexturePicker(void)
{
	cv_gtk_gl_texpick_layout=Cfg_DefVar(config_fe,
		"gtk_gl.texpick.layout","0",
		"Texture layout method for the OpenGL/Gtk+ texture picker.\n"
		"0=scroll left-right, alphabetical order\n"
		"1=scroll left-right, squeezed together\n"
		"2=scroll up-down, alphabetical order\n"
		"3=scroll up-down, squeezed together\n");
      
	cv_gtk_gl_texpick_scale=Cfg_DefVar(config_fe,
		"gtk_gl.texpick.scale","1",
		"Default scale for textures in the OpenGL/Gtk+ texture picker.");
      
	cv_gtk_gl_texpick_min_size=Cfg_DefVar(config_fe,
		"gtk_gl.texpick.min_size","0",
		"Minimum size of textures in the texture picker (0 for no "
		"minimum). Textures smaller than this will be enlarged.");
      
	cv_gtk_gl_texpick_max_size=Cfg_DefVar(config_fe,
		"gtk_gl.texpick.max_size","128",
		"Maximum size of textures in the texture picker (0 for no "
		"maximum, not recommended). Textures larger than this will be "
		"shrunk");
}

