#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "brush.h"

#include "error.h"
#include "game.h"
#include "geom.h"
#include "map.h"
#include "memory.h"
#include "quest.h"
#include "texalign.h"
#include "uid.h"
#include "undo.h"
#include "fe/render.h"


static int Q3C_Copy(brush_t *dst,const brush_t *src)
{
   int size;
   
   dst->x.q3c=src->x.q3c;

   /* size of texture alignment information */
   size=src->x.q3c.sizex*src->x.q3c.sizey;
   
   if (src->x.q3c.a)
   {
      dst->x.q3c.a=Q_malloc(size*sizeof(vec2_t));
      if (!dst->x.q3c.a)
		{
			ERR_AMEM();
         return 0;
		}
      memcpy(dst->x.q3c.a,src->x.q3c.a,size*sizeof(vec2_t));
   }
   return 1;
}

static void Q3C_Free(brush_t *b)
{
   if (b->x.q3c.a)
      Q_free(b->x.q3c.a);
   b->x.q3c.a=NULL;
}


static int N_Copy(brush_t *dst,const brush_t *src)
{
   int i;
   face_t **d,**s;

   for (i=src->num_faces,d=dst->faces,s=src->faces;i>0;i--,s++,d++)
      (*d)->a=(*s)->a;
   /* texv will be taken care of by the update function */
   return 1;
}

static void N_Free(brush_t *b)
{
   int i;
   face_t **f;

   for (i=b->num_faces,f=b->faces;i>0;i--,f++)
      if ((*f)->texv)
         Q_free((*f)->texv);
}

static void N_Update(brush_t *b)
{
   int i;
   for (i=0;i<b->num_faces;i++)
      CalcTexv(b,i);
   strcpy(b->tex.name,b->num_faces?b->faces[0]->tex.name:DEFAULT_TEXTURE);
}


brush_type_t brush_types[]=
{
{BRUSH_NORMAL  ,"normal brush"       ,BRUSH_F_NORMAL|BRUSH_F_CSG,
   NULL,N_Copy,N_Free,N_Update},
{BRUSH_SUBTRACT,"subtractive brush"  ,BRUSH_F_NORMAL,
   NULL,N_Copy,N_Free,N_Update},
{BRUSH_POLYGON ,"single polygon"     ,BRUSH_F_NORMAL,
   NULL,N_Copy,N_Free,N_Update},

{BRUSH_Q3_CURVE,"Quake 3 curve"      ,0,NULL,Q3C_Copy,Q3C_Free},

{0}
};

static brush_type_t *B_LookupType(int type)
{
   brush_type_t *bt;
   
   for (bt=brush_types;bt->type;bt++)
      if (bt->type==type)
         return bt;

	ERROR_ADD("Unknown brush type %i!",type);
   return NULL;
}


brush_t *B_New(int type)
{
   brush_t *b;

	b=Mem_Alloc(&mem_brush);
   if (!b)
	{
		ERR_AMEM();
		return NULL;
	}

   b->bt=B_LookupType(type);
	if (!b->bt)
	{
		Mem_Free(&mem_brush,b);
		return NULL;
	}
   if (b->bt->b_new)
   {
      if (!b->bt->b_new(b))
      {
         B_Free(b);
         return NULL;
      }
   }

   return b;
}

void B_FreeC(brush_t *b)
{
   int i;
   face_t **f;

   ASSERT(!b->real || (!b->Next && !b->Last));

   if (b->bt->b_free) b->bt->b_free(b);

#define FREE(x) if (x) { Q_free((x)); (x)=NULL; }
   
   if (b->faces)
   {
      for (i=0,f=b->faces;i<b->num_faces;i++,f++)
      {
         S_R_FaceFree(*f);
         G_Texdef_Free(&(*f)->tex);
         FREE((*f)->verts)
			Mem_Free(&mem_face,*f);
      }
      FREE(b->faces);
   }
   
   S_R_BrushFree(b);

   FREE(b->verts)
   
   FREE(b->edges)

#undef FREE
}

void B_Free(brush_t *b)
{
   B_FreeC(b);
	Mem_Free(&mem_brush,b);
}


void B_Link(brush_t *b)
{
   ASSERT(!b->Next && !b->Last);

   if (M.BrushHead!=NULL) M.BrushHead->Last = b;
   b->Next = M.BrushHead;
   b->Last = NULL;
   M.BrushHead = b;
   b->real=1;
   b->VNext=b->VLast=NULL;

   b->uid=GetBrushUID();

   M.num_brushes++;

   U_BrushAdd(b);

   B_Update(b);
}

void B_Unlink(brush_t *b)
{
   ASSERT(b->Next || b->Last || b==M.BrushHead);
   ASSERT(b->real);

   U_BrushDelete(b);

   M.num_brushes--;
   
   if (b->Next != NULL) b->Next->Last = b->Last;
   if (b->Last != NULL) b->Last->Next = b->Next;
   if (b==M.BrushHead) M.BrushHead=b->Next;

   b->Next=NULL;
   b->Last=NULL;
   b->VNext=b->VLast=NULL;
   b->real=0;
}


void B_ChangeType(brush_t *b,int ntype)
{
   b->bt=B_LookupType(ntype);
}


brush_t *B_Duplicate(brush_t *dest, const brush_t *source)
{
	int i;

   if (!dest)
   {
		dest=Mem_Alloc(&mem_brush);
      if (!dest)
      {
			ERR_AMEM();
         return NULL;
      }
   }
	else
	   memset(dest,0,mem_brush.esize);
   
	/* Copy face Information */
	dest->faces = (face_t **)Q_malloc(sizeof(face_t *) * (source->num_faces));
	if (dest->faces==NULL)
	{
		ERR_AMEM();
		return NULL;
	}
   memset(dest->faces,0,sizeof(face_t *)*source->num_faces);
	dest->num_faces = source->num_faces;

	for (i=0;i<(source->num_faces);i++)
	{
		dest->faces[i]=Mem_Alloc(&mem_face);
		if (!dest->faces[i])
		{
			ERR_AMEM();
			return NULL;
		}
		/* Copy Vertex Information from the face */
		dest->faces[i]->verts = (int *)Q_malloc(sizeof(int) * (source->faces[i]->num_verts));
		if (dest->faces[i]->verts==NULL)
		{
			ERR_AMEM();
			return NULL;
		}
      memcpy(dest->faces[i]->verts,source->faces[i]->verts, sizeof(int) * (source->faces[i]->num_verts) );
		dest->faces[i]->num_verts = source->faces[i]->num_verts;

      dest->faces[i]->normal=source->faces[i]->normal;
      strcpy(dest->faces[i]->tex.name,source->faces[i]->tex.name);
      if (!G_Texdef_Copy(&dest->faces[i]->tex,&source->faces[i]->tex))
      {
			ERR_AMEM();
         return NULL;
      }
	}

   strcpy(dest->tex.name,source->tex.name);
   if (!G_Texdef_Copy(&dest->tex,&source->tex))
   {
		ERR_AMEM();
      return NULL;
   }

	dest->num_verts = source->num_verts;
   
	dest->verts = (vec3_t *)Q_malloc(sizeof(vec3_t) * (source->num_verts));
	if (dest->verts==NULL)
	{
		ERR_AMEM();
		return NULL;
	}
   memcpy(dest->verts,source->verts,sizeof(vec3_t) * (source->num_verts));
   
	dest->num_edges = source->num_edges;
   if (source->num_edges)
   {
   	dest->edges = (edge_t *)Q_malloc(sizeof(edge_t) * (source->num_edges));
   	if (dest->edges==NULL)
   	{
			ERR_AMEM();
   		return NULL;
   	}
   	
   	for (i=0;i<(source->num_edges);i++)
   	{
   		dest->edges[i] = source->edges[i];
   	}
   }
	
	dest->Entity = source->Entity;
	dest->Group = source->Group;

   dest->bt=source->bt;

   if (dest->bt->b_copy)
      dest->bt->b_copy(dest,source);

   return dest;
}


int B_Size(const brush_t *b)
{
   int s;
   int i;
   face_t **f;

/* TODO: add this to get accurate sizes? worth it? */
/*    if (b->bt->b_size) b->bt->b_size(b);*/

   s=0;
   if (b->faces)
   {
      for (i=0,f=b->faces;i<b->num_faces;i++,f++)
      {
         s+=sizeof(int)*(*f)->num_verts;
      }
      s+=(mem_face.esize+sizeof(face_t *))*b->num_faces;
   }

   if (b->verts) s+=sizeof(vec3_t)*b->num_verts;
   if (b->edges) s+=sizeof(edge_t)*b->num_edges;

   return s;
}


void B_Update(brush_t *b)
{
   int i;

   CalcBrushCenter(b);
   RecalcNormals(b);

   if (b->bt->b_update)
      b->bt->b_update(b);

   for (i=0;i<b->num_faces;i++)
      S_R_FaceUpdate(b->faces[i]);
   S_R_BrushUpdate(b);
}


face_t *F_Copy(face_t *dst,const face_t *src)
{
   if (!dst)
   {
      dst=Mem_Alloc(&mem_face);
      if (!dst) return NULL;
   }

   *dst=*src;
   
   dst->verts=Q_malloc(sizeof(int)*src->num_verts);
   if (!dst->verts) return NULL;
   memcpy(dst->verts,src->verts,sizeof(int)*src->num_verts);

   dst->tex.g=NULL;
   if (!G_Texdef_Copy(&dst->tex,&src->tex))
      return NULL;

   dst->texv=NULL;
   
   return dst;
}

face_t *F_CopyP(face_t *dst,const face_t *src)
{
   if (!dst)
   {
      dst=Mem_Alloc(&mem_face);
      if (!dst) return NULL;
   }

   *dst=*src;

   dst->verts=NULL;
   dst->num_verts=0;

   dst->tex.g=NULL;
   if (!G_Texdef_Copy(&dst->tex,&src->tex))
      return NULL;

   dst->texv=NULL;
   
   return dst;
}


int B_AllocFaces(brush_t *b,int num_faces)
{
	face_t **fp;
	int i;
	fp=Q_malloc(sizeof(face_t *)*num_faces);
	if (!fp) return 0;
	b->faces=fp;
	for (i=0;i<num_faces;i++)
	{
		b->faces[i]=Mem_Alloc(&mem_face);
		if (!b->faces[i]) return 0;
	}
	b->num_faces=num_faces;
	return 1;
}

